from obspy import read, read_inventory
import spectral as sp
import sys

## Set variables
# DOCTAR data
datafile_base='../0_Data/DOCTAR/D01.DO.???..D.2011.236.000000';
metafile='../0_Data/DOCTAR/DOCTAR_inventory.xml';
length=12*3600; # 12 hours, avoids EQ at 2011-08-24T18
drivechan='3'
respchan='4'
noisechan='equal'
# Marianas data
datafile_base='../Marianas/XF.B12.2012_10_01-05_2Hz.mseed';
metafile='../Marianas/XF.B12.station.xml';
length=None  # read all
drivechan='1'
respchan='Z'
noisechan='response'

def main():
    ## Read in data and responses
    print('Reading channels ',end='')
    st=read(datafile_base)
    for tr in st:
        print(tr.stats.channel,end=', ')
    print('and prepping meta/data')
    inv=read_inventory(metafile)
    st.attach_response(inv)
    # Cut data to have start at same time on all channels
    dt=st[0].stats.starttime+60
    if length:
        st.trim(dt,dt+length)

    ## Calculate spectra and transfer functions
    print('Calculating PSDs...')
    s=sp.calc_PSDs(st,window_length=1000);
    print('Calculating coherences...')
    c=sp.calc_cohers(st,window_length=1000);
    print('Calculating transfer functions...')
    x=sp.calc_XF(s,c,drivechan,respchan,noisechan)

    ## Plot all
    print('Plotting...')
    header='PLOTS/{}.{}'.format(st[0].stats.network,st[0].stats.station)
    st.plot(size=(800,800),outfile='{}_time_obspy.png'.format(header))
    sp.plot_spects(s,outfile='{}_spect_spectral.png'.format(header))
    sp.plot_cohers(c,outfile='{}_coher_spectral.png'.format(header))
    sp.plot_XF(x,outfile='{}_XF_spectral.png'.format(header))

    ## Plot PPSD
#     from obspy.signal import PPSD
#     ppsd = PPSD(st[1].stats, inv)
#     print(ppsd.id)
#     ppsd.add(st[1]) 
#     ppsd.plot() 

    
if __name__ == "__main__": main()


"""
Seismology Data Preparation Module

Routines and classes providing a consistent interface for Python programs using
the SDPCHAIN protocol, which includes:

- Create or append to a process-steps.json file at each step
    - the process-steps file is read from the input directory and written
      to the output directory.  If both directories are the same the new step
      is appended to the existing file
- command-line arguments:
    - -d (base directory)
    - -i (onput directory), can be relative to base directory
    - -o (output directory), can be relative to base directory
    - optional (input_file) or (input_files), will have input directory pre-pended
    - optional -of (output_file) or -ofs (output_files), will have output dir pre-pended
    - optional -f, forces writing of output file if it already exists
  
Classes
---------------------

:ProcessSteps: Holds process-step information and saves/appends to file

NOT YET VERIFIED
:ArgParser: argparse:argparser instance prefilled with the SDPCHAIN command-
            line arguments.  Once parsed, the optional infile, infiles, outfile
            and outfiles attributes are adusted for their relation to the
            base and input/output directories and any process-steps.json file
            in the input directory is copied to the base directory (quits
            if there is already one there)

Command-line Routines
---------------------

These routines perform common functions while following the
SDPCHAIN rules (process-steps file, -i, -o, -d)
:sdpstep: run a standard command-line program 
:sdpcat: concatenate binary files
"""
name = "sdpchain"
from .process_steps import ProcessStep
from .sdpcat import sdpcat
from .sdpstep import main as sdpstep, is_tool
from .version import __version__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Dump record information from an LCHEAPO file
"""
from pathlib import Path

import sys
import argparse
import datetime as dt

from .lcheapo_utils import (LCDiskHeader, LCDirEntry, LCDataBlock)


# ------------------------------------
# Global Variable Declarations
# ------------------------------------
PROGRAM_NAME = "lcdump"
VERSION = "0.2.3"
version_notes = """
    v0.2 (2015/01): WCC added format options
    v0.2.1 (2017/01): WCC added possibility to compare time with theoretical
    v0.2.2 (2017/03): WCC added directory printing
    v0.2.3 (2017/03): Added "--from_end" option
    """


def getOptions():
    """
    Parse command line arguments
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("inFilename", help="LCHEAPO file name")
    parser.add_argument("startBlock", type=int, default=5000,
                        help="block at which to start dump")
    parser.add_argument("nBlocks", type=int, default=10,
                        help="number of blocks to dump")
    parser.add_argument("-r", "--from_end", default=False, action="store_true",
                        help="Count startBlocks back from end of file")
    parser.add_argument("-v", "--version", default=False, action="store_true",
                        help="Display Program Version")
    parser.add_argument("-p", "--printHeader", default=False,
                        action="store_true",  help="Print disk header")
    parser.add_argument("-d", "--printDirectory", default=False,
                        action="store_true",  help="Print disk directory")
    parser.add_argument("-f", "--format", type=int, default=0,
                        choices=[0, 1, 2, 3],
                        help="Output format: 0=pretty [default], 1=decimal,\
                        2=hex, 3=time_verify")
    args = parser.parse_args()

    # Get the filename (the arguments)
    if args.version:
        print("{} - Version: {}".format(PROGRAM_NAME, VERSION))
        sys.exit(0)

    return args


def main():

    args = getOptions()

    if args.from_end is True:
        args.startBlock = int(Path(args.inFilename).stat().st_size/512)-args.startBlock

    fp = open(args.inFilename, 'rb')
    if args.printHeader or (args.format == 3) or args.printDirectory:
        lcHeader = LCDiskHeader()
        lcHeader.readHeader(fp)
        if args.printHeader:
            lcHeader.printHeader()
        if args.printDirectory:
            lcDir = LCDirEntry()
            lcDir.seekBlock(fp, lcHeader.dirStart)
            iDir = 0
            print("="*80)
            print(" {:25s} {:8s} {:15s} {:10s} {:12s} {:11s} {}".format(
                "DateTime", "MuxChan", "SPS", "block#",
                "numBlocks", "recordLen", "Flag"))
            print("="*80)
            while iDir < lcHeader.dirCount:
                lcDir.readDirEntry(fp)
                print(lcDir)
                iDir = iDir + 1

    if args.format == 3:
        firstBlock = lcHeader.dataStart
        if firstBlock == 0:   # normal start block
            firstBlock = 2393
        lcStartData = LCDataBlock()
        lcStartData.seekBlock(fp, firstBlock)
        lcStartData.readBlock(fp)
        firstTime = dt.datetime(lcStartData.year+2000,
                                lcStartData.month,
                                lcStartData.day,
                                lcStartData.hour,
                                lcStartData.minute,
                                lcStartData.second,
                                lcStartData.msec*1000)
        print("First block = {:d}: Time = {} ".format(firstBlock,
              firstTime.strftime('%Y/%m/%d-%H:%M:%S.%f')[:-3]))
        sampRate = lcHeader.realSampleRate
        if sampRate == 0:
            sampRate = 62.5
            print("Sample rate not in directory header, assuming 62.5")
        print(" {:>7s}: {:^2s} | {:^23s} | {:^23s} | {:>8s} ".format(
              "BLOCK", "CH", "EXPECTED TIME", "FOUND TIME", "DELTA"))
        print("-{:->7s}:-{:-^2s}-|-{:-^23s}-|-{:-^23s}-|-{:->8s}-".format(
              "-", "-", "-", "-", "-"))

    block = args.startBlock
    lcData = LCDataBlock()
    lcData.seekBlock(fp, block)

    for i in range(0, args.nBlocks):
        lcData.readBlock(fp)
        print("{:8d}:".format(args.startBlock + i), end=' ')
        if args.format == 3:
            time = dt.datetime(lcData.year + 2000,
                               lcData.month,
                               lcData.day,
                               lcData.hour,
                               lcData.minute,
                               lcData.second,
                               lcData.msec * 1000)
            rec_offset = int((args.startBlock + i - firstBlock) / 4)
            calcTime = firstTime + dt.timedelta(
                       seconds=rec_offset*lcData.numberOfSamples/sampRate)
            print("{:2d} | {} | {} | {:8.3f}".format(
                lcData.muxChannel,
                calcTime.strftime('%Y/%m/%d-%H:%M:%S.%f')[:-3],
                time.strftime('%Y/%m/%d-%H:%M:%S.%f')[:-3],
                (time-calcTime).total_seconds())
            )
        elif args.format == 1:
            lcData.printDecimalDumpOfHeader(True)
        elif args.format == 2:
            lcData.printHexDumpOfData()
        elif args.format == 0:
            lcData.prettyPrintHeader()
        else:
            print("ERROR! Shouldn't get here!")


# Run 'main' if the script is not imported as a module
if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Cut an LCHEAPO file into pieces

Used to remove bad/empty blocks, blocks start with 0 and are 512-bytes
"""
import argparse
import copy
# import os
from datetime import datetime as dt
import sys
from math import floor
from pathlib import Path

from .sdpchain import ProcessStep
from .version import __version__

BLOCK_SIZE = 512
MAX_BLOCK_READ = 2048   # max number of blocks to read at once


def main():
    return_code = 0
    exec_messages = []

    # GET ARGUMENTS
    args = getOptions()

    # Verify output filename
    if not args.output_file:
        # Create output filename
        base = Path(args.input_files[0]).stem
        ext = Path(args.input_files[0]).suffix
        args.output_file = f'{base}_{args.start:d}_{args.end:d}{ext}'
    out_path = Path(args.out_dir) / args.output_file
    if out_path.exists():
        print(f'output file {str(out_path)} exists already, quitting...')
        sys.exit(2)

    with open(Path(args.in_dir) / args.input_files[0], 'rb') as fp:
        # Set/validate last block to read
        fp.seek(0, 2)   # End of file
        last_file_block = floor(fp.tell()/BLOCK_SIZE)-1
        if args.end:
            if args.end > last_file_block:
                args.end = last_file_block
        else:
            args.end = last_file_block
        # Quit if start block is after EOF and/or end block
        if args.start > last_file_block:
            return_code = 2
            msg = 'Error: --start block [{:d}] is beyond EOF [{:d}]'.format(
                args.start, last_file_block)
            print(msg)
            exec_messages.append(msg)
        elif args.start > args.end:
            return_code = 3
            msg = 'Error: --start block [{:d}] is beyond --end [{:d}]'.format(
                args.start, args.end)
            print(msg)
            exec_messages.append(msg)
        else:
            msg = 'Writing blocks {:d}-{:d} to {}'.format(
                args.start, args.end, args.output_file)
            if args.quiet is False:
                print(msg)
            exec_messages.append(msg)
            with open(out_path, 'wb') as of:
                start_block = args.start
                fp.seek(start_block * BLOCK_SIZE, 0)
                while start_block <= args.end:
                    if (args.end-start_block+1) < MAX_BLOCK_READ:  # NEAR EOF
                        of.write(fp.read((args.end
                                          - start_block
                                          + 1) * BLOCK_SIZE))
                        break
                    else:
                        of.write(fp.read(MAX_BLOCK_READ * BLOCK_SIZE))
                        start_block += MAX_BLOCK_READ
    # Save/append process information to process_steps file
    global process_step
    process_step.messages = exec_messages
    process_step.out_file = args.output_file
    process_step.exit_code = return_code
    process_step.write(args.in_dir, args.out_dir, verbose=True)


def getOptions():
    """
    Parse user passed options and parameters.
    """
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("input_files", nargs=1, help="Input filename")
    parser.add_argument("--of", dest="output_file", default="",
                        help="""
                        Output filename.  If not provided, writes to
                        {root}_{start}_{end}.{ext}, where {ext} is the last
                        pathname component of {in_file_name} and {root} is
                        the rest""")
    parser.add_argument("--start", type=int, default=0,
                        help="start block to write out (0 if not specified)")
    parser.add_argument("--end", type=int, default=0,
                        help=""" last block to write out (end of file if not
                        specified)""")
    parser.add_argument("-d", "--directory", dest="base_dir",
                        default='.', help="Base directory for files")
    parser.add_argument("-i", "--input", dest="in_dir", default='.',
                        help="path to input files (abs, or rel to base)")
    parser.add_argument("-o", "--output", dest="out_dir", default='.',
                        help="path for output files (abs, or rel to base)")
    parser.add_argument("--quiet", action='store_true',
                        help="Don't print to stdout for normal operation")
    args = parser.parse_args()
    global process_step
    process_step = ProcessStep('lccut',
                               " ".join(sys.argv),
                               app_description=__doc__,
                               app_version=__version__,
                               parameters=args)
    args.in_dir, args.out_dir, _ = ProcessStep.setup_paths(args)
    return args


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Make a directory containing lcheapo examples
import argparse
import shutil
import os
import inspect


def main():
    """
    Copy _examples file to a local directory
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-d", "--destdir", default='EXAMPLES_LCHEAPO_OBSPY',
                        help="destination directory",)
    args = parser.parse_args()

    self_path = os.path.dirname(os.path.abspath(inspect.getfile(
        inspect.currentframe())))
    examples_dir = os.path.join(self_path, '_examples')

    if os.path.exists(args.destdir):
        print(f'destination directory "{args.destdir}" already exists')
    else:
        try:
            shutil.copytree(examples_dir, args.destdir)
        except shutil.Error as e:
            print(f'Directory not copied. Error: {e}')
        except OSError as e:
            print(f'Directory not copied. Error: {e}')


# ---------------------------------------------------------------------------
# Run 'main' if the script is not imported as a module
# ---------------------------------------------------------------------------
# if __name__ == '__main__':
#     main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Create/modify an LCHEAPO data file header
"""
import sys
import copy
# import os
import math as m
import argparse
from datetime import datetime, timedelta
from pathlib import Path

from .lcheapo_utils import (LCDiskHeader, LCDirEntry)
from .version import __version__

from .sdpchain.process_steps import ProcessStep

# ------------------------------------
# Global Variable Declarations
# ------------------------------------
PROGRAM_NAME = "lcheader"
bytes_per_block = 512
dirEntries_per_dirBlock = 16
blocks_per_dirEntry = 14336
samples_per_block = 166
accepted_sample_rates = [31.25, 62.5, 125, 250, 500, 1000, 2000, 4000]
accepted_num_channels = [1, 2, 3, 4]


def main():
    """main function"""
    opts = __getOptions()
    
    params = dict(wake_time=datetime(2017, 1, 1, 5, 0, 0),
                  end_time=datetime(2018, 1, 1, 5, 0, 0),
                  output_filename='generic.header.lch')
    if opts.input_file:
        h = __read_header(Path(opts.in_dir) / opts.input_file)
    else:
        h = __generic_header()
    if opts.description:
        h.description = opts.description
    if opts.sample_rate:
        h.realSampleRate = opts.sample_rate
        h.sampleRate = m.floor(opts.sample_rate)
    if opts.num_channels:
        h.numberOfChannels = opts.num_channels
    if opts.wake_time:
        params['wake_time'] = __get_datetime(None, opts.wake_time)
    if opts.end_time:
        params['end_time'] = __get_datetime(None, opts.end_time)
    if opts.output_file:
        params['output_filename'] = opts.output_file
    h, params = _modify_parameters(h, params, opts)
    if opts.dry_run:
        sys.exit(2)
    with open(Path(opts.out_dir) / params['output_filename'], 'wb') as fp:
        # Pre-blank file
        if opts.noDirectory:
            fp.write(b'\x00' * bytes_per_block * h.dirStart)
        else:
            fp.write(b'\x00' * bytes_per_block * h.dataStart)

        # Write header
        h.seekHeaderPosition(fp)
        h.writeHeader(fp)
        if opts.noDirectory:
            sys.exit(0)

        # Write directory
        d = __prep_dirEntry(h.sampleRate)
        dt = params['wake_time']
        dirCount = 0
        samples_per_dirEntry = samples_per_block * blocks_per_dirEntry / \
            h.numberOfChannels
        dir_timeoffset = timedelta(seconds=samples_per_dirEntry / h.sampleRate)
        blocknum = h.dataStart
        d.seekBlock(fp, h.dirStart)
        while dt < params['end_time']:
            d.blockNumber = blocknum
            d.changeTime(dt)
            d.writeDirEntry(fp)
            dt += dir_timeoffset
            dirCount += 1
            blocknum += blocks_per_dirEntry
        h.dirCount = dirCount
        h.dirBlock = h.dirStart + int(dirCount/dirEntries_per_dirBlock)
        h.seekHeaderPosition(fp)
        h.writeHeader(fp)
        
    global process_step
    process_step.parameters['wake_time'] = params['wake_time'].isoformat()
    process_step.parameters['end_time'] = params['end_time'].isoformat()
    process_step.output_file = params['output_filename']
    process_step.exit_status = 0
    process_step.write(opts.in_dir, opts.out_dir) 
    sys.exit(0)


def __getOptions():
    """
    Parse user passed options and parameters
    """
    # usageStr = "%prog [-h] [-V]"
    parser = argparse.ArgumentParser(
        description="Create/modify an LCHEAPO data file header",
        epilog="Interactive unless you specify --description, -s, -c, -w, -e "
               "and --output_file")
    parser.add_argument("-V", "--version", default=False, action="store_true",
                        help="Display Program Version")
    parser.add_argument("-n", "--noDirectory", default=False,
                        action="store_true", help="don't create a directory")
    parser.add_argument("--dry_run", default=False, action="store_true",
                        help="don't save to disk")
    parser.add_argument("--no_questions", default=False, action="store_true",
                        help="make a generic header, no questions", )
    parser.add_argument("--input_file",
                        help="input header file to modify")
    parser.add_argument("--description",
                        help="Header Description field (e.g. "
                             "Experiment_OBSType_SN_Station)")
    parser.add_argument("-s", "--sample_rate", type=float, metavar="SAMP_RATE",
                        choices=accepted_sample_rates,
                        help=f"Sample Rate. Allowed = {accepted_sample_rates}")
    parser.add_argument("-c", "--num_channels", type=int, metavar="NCHANS",
                        choices=accepted_num_channels,
                        help=f"Number of channels. Allowed = {accepted_num_channels}")
    parser.add_argument("-w", "--wake_time",  metavar="TIME",
                        help="Start of recorded data (ISO0601 format)")
    parser.add_argument("-e", "--end_time",  metavar="TIME",
                        help="End of recorded data (ISO0601 format)")
    parser.add_argument("--output_file", help="Output filename")
    parser.add_argument("-d", "--directory", dest="base_dir",
                        default='.', help="Base directory for files")
    parser.add_argument("-i", "--input", dest="in_dir", default='.',
                        help="path to input files (abs, or rel to base)")
    parser.add_argument("-o", "--output", dest="out_dir", default='.',
                        help="path for output files (abs, or rel to base)")
    args = parser.parse_args()
    global process_step
    process_step = ProcessStep('lcheader',
                               " ".join(sys.argv),
                               app_description='Create LCHEAPO header and '
                                               'directory',
                               app_version=__version__,
                               parameters=args)

    args.input_files = [args.input_file]
    args.in_dir, args.out_dir, _ = ProcessStep.setup_paths(args, expand_wildcards=False)
    args = parser.parse_args()

    # Get the filename (the arguments)
    if args.version:
        print(f"{PROGRAM_NAME} - Version: {__version__}")
        sys.exit(0)
    return args


def _modify_parameters(h, params, opts):
    """Imitate LCHEAPO parameter menu, return values"""
    filename_prefix = params['output_filename'].split('.')[0]
    while not __validate_params(h, params, opts):
        h.description = __get_string("Description (Cruise_InstModel_SN_Site)",
                                     h.description)
        h.realSampleRate = __get_float("Sample Rate", h.realSampleRate,
                                       accepted_sample_rates)
        h.sampleRate = m.floor(h.realSampleRate)
        h.numberOfChannels = __get_int("Number of Channels",
                                       h.numberOfChannels,
                                       accepted_num_channels)

        params['wake_time'] = __get_datetime("Wake Time", params['wake_time'])
        if params['end_time'] < params['wake_time']:
            params['end_time'] = params['wake_time'] + timedelta(days=365)
        params['end_time'] = __get_datetime("End Time", params['end_time'])
        filename_prefix = __get_string("Output Filename Prefix",
                                       filename_prefix)
        params['output_filename'] = filename_prefix + '.header.lch'

    datalen = params['end_time'] - params['wake_time']
    datalen_seconds = datalen.days * 86400 + datalen.seconds + \
        datalen.microseconds / 1.e6
    datalen_blocks = int(datalen_seconds * h.sampleRate / samples_per_block) *\
        h.numberOfChannels
    h.writeBlock = h.dataStart + datalen_blocks

    params['sample_rate'] = h.realSampleRate
    params['number_of_channels'] = h.numberOfChannels
    params['description'] = h.description

    return h, params


def __validate_params(h, params, opts):
    """Validate header parameters"""
    if opts.no_questions:
        return True

    print('**** PARAMETERS ****')
    print(f'         Description: {h.description}')
    print(f'         Sample Rate: {h.realSampleRate:g}')
    print(f'  Number of Channels: {h.numberOfChannels}')
    print(f'           Wake Time: {params["wake_time"].isoformat()}')
    print(f'            End Time: {params["end_time"].isoformat()}')
    print(f'     Output Filename: {params["output_filename"]}')

    if opts.description and opts.sample_rate and opts.num_channels \
            and opts.wake_time and opts.end_time and opts.output_file:
        return True

    resp = input('Is this acceptable? [y/N]: ')
    if not resp:
        return False
    if resp[0].lower() == 'y':
        return True
    return False


def __read_header(filename):
    """Read in header from an LCHEAPO file"""
    h = LCDiskHeader()
    with open(filename, 'r') as fp:
        h.seekHeaderPosition(fp)
        h.readHeader(fp)
    return h


def __generic_header():
    """Create a generic lcheapo header"""

    default_data_startblock = 3586
    default_dir_startblock = 8
    default_data_type = 2
    default_data_type_string = "Uncompressed (24-Bit)"

    h = LCDiskHeader()

    # UNUSED OR STANDARD VALUES
    h.dataType = default_data_type
    h.diskSize = 0    # GB
    h.ramSize = 8     # MB
    h.softwareVersion = '9.08a-OLD'
    h.dirStart = default_dir_startblock
    h.dataStart = default_data_startblock
    # h.dirSize=(h.dataStart-h.dirStart-1)*dirEntries_per_dirBlock
    # Modified to work with lc2ms v1
    h.dirSize = (h.dataStart - h.dirStart) * dirEntries_per_dirBlock - 1
    (h.slowStart, h.slowSize, h.slowBlock, h.slowByte) = (0, 0, 0, 0)
    (h.logStart, h.logSize, h.logBlock, h.logByte) = (0, 0, 0, 0)
    (h.slowDataRate, h.slowStartChannel, h.slowNumberOfChannels) = (0, 0, 0)
    (h.readBlock, h.startChannel, h.diskNumber) = (0, 0, 0)
    (h.numberOfWindows, h.readByte, h.writeByte) = (0, 0, 0)

    # These should be modified after writing directory
    h.dirBlock, h.dirCount = (0, 0)

    # USER-SPECIFIED VALUES
    h.description = 'generic lcheapo header'
    h.sampleRate = 125
    h.numberOfChannels = 4
    h.writeBlock = h.dataStart

    # Additions which cannot be written to file
    h.realSampleRate = h.getRealSampleRate(h.sampleRate)
    h.dataTypeString = default_data_type_string
    return h


def __prep_dirEntry(sample_rate):
    """Prepare lcheapo directory entry object"""
    d = LCDirEntry()
    d.recordLength = 0  # Unused
    d.sampleRate = sample_rate
    d.numBlocks = blocks_per_dirEntry
    d.flag = 0x49
    d.muxChannel = 0
    d.U1 = 0            # Unused
    return d


def __get_string(question, default):
    """Read a string from the command line"""
    resp = input("{} [{}]: ".format(question, default))
    if not resp:
        return default
    return resp


def __get_datetime(question, default):
    """Read a datetime from the command line"""
    if question is not None:
        resp = input("{} [{}]: ".format(question, default.isoformat()))
        if not resp:
            return default
    else:
        resp = default
    try:
        resp = datetime.strptime(resp, '%Y-%m-%dT%H:%M:%S')
    except ValueError:
        try:
            resp = datetime.strptime(resp, '%Y-%m-%dT%H:%M:%S.%f')
        except ValueError:
            print('Invalid input: {}'.format(resp))
            if question is not None:
                resp = __get_datetime(question, default)
            else:
                sys.exit(2)
    return resp


def __get_int(question, default, possibles=False):
    """Read an int from the command line"""
    resp = input("{} [{:d}]: ".format(question, default))
    if not resp:
        return default
    else:
        try:
            resp = int(resp)
        except ValueError:
            print('Invalid input: {}'.format(resp))
            resp = __get_int(question, default, possibles)
        if isinstance(possibles, list):
            if resp not in possibles:
                print('Input must be one of: {}'.format(str(possibles)))
                resp = __get_int(question, default, possibles)
    return resp


def __get_float(question, default, possibles=False):
    """Read a float from the command line"""
    resp = input("{} [{:g}]: ".format(question, default))
    if not resp:
        return default
    else:
        try:
            resp = float(resp)
        except ValueError:
            if isinstance(possibles, list):
                print('Input must be one of: {}'.format(str(possibles)))
            else:
                print('Invalid input: {}'.format(resp))
            resp = __get_float(question, default, possibles)
        if isinstance(possibles, list):
            if resp not in possibles:
                print('Input must be one of: {}'.format(str(possibles)))
                resp = __get_float(question, default, possibles)
    return resp


# ---------------------------------------------------------------------------
# Run 'main' if the script is not imported as a module
# ---------------------------------------------------------------------------
if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
create miniSEED file(s) from LCHEAPO file(s)
"""
import argparse
# import os
import sys
import datetime
import inspect
from pathlib import Path

from .sdpchain import ProcessStep

from .chan_maps import chan_maps
from .lcread import read as lcread
from .version import __version__


def lc2ms():
    """
    Convert fixed LCHEAPO data to basic miniSEED files

    NO drift or leapsecond correction:
    """
    print(lc2ms.__doc__)
    parser = argparse.ArgumentParser(
        description=inspect.cleandoc(lc2ms.__doc__),
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("infile", help="Input filename")
    parser.add_argument("-t", "--obs_type", default='SPOBS2',
                        help="obs type.  Controls channel and location codes",
                        choices=[s for s in chan_maps])
    parser.add_argument("--station", default='SSSSS',
                        help="station code for this instrument")
    parser.add_argument("--network", default='XX',
                        help="network code for this instrument")
    parser.add_argument("-d", dest="base_dir", metavar="BASE_DIR",
                        default='.', help="base directory for files")
    parser.add_argument("-i", dest="in_dir", metavar="IN_DIR", default='.',
                        help="input file directory (absolute, " +
                             "or relative to base_dir)")
    parser.add_argument("-o", dest="out_dir", metavar="OUT_DIR", default='.',
                        help="output file directory (absolute, " +
                             "or relative to base_dir)")
    parser.add_argument("-v", "--verbose", action='store_true',
                        help="verbose output")
    parser.add_argument("--version", action='store_true',
                        help="Print version number and quit")
    args = parser.parse_args()
    parameters = vars(args).copy()
    if args.version is True:
        print(f"Version {__version__}")
        sys.exit(0)

    # ADJUST INPUT PARAMETERS
    args.input_files = [args.infile]
    process_step = ProcessStep('lc2ms_weak',
                               " ".join(sys.argv),
                               app_description=__doc__,
                               app_version=__version__,
                               parameters=parameters)
    args.in_dir, args.out_dir, args.infiles = ProcessStep.setup_paths(args)
    # Expand captured wildcards
    # args.infiles = [x.name for f in args.infiles
    #                 for x in Path(args.in_dir).glob(f)]

    stream = lcread(Path(args.in_dir) / args.infile, network=args.network,
                    station=args.station, obs_type=args.obs_type,
                    starttime=0,
                    endtime = 1*86400*365.25) # For up to 1 year of data
    out_dir = Path(args.out_dir)
    out_dir.mkdir(parents=True, exist_ok=True)
    out_files = []
    for tr in stream:
        s = tr.stats
        out_files.append(f'{s.network}.{s.station}.{s.location}.{s.channel}.mseed')
        fname = str(out_dir / out_files[-1])
        tr.write(fname, format='MSEED', encoding='STEIM1', reclen=4096)
    return_code = 0
    process_step.output_files = out_files
    process_step.exit_code = return_code
    process_step.write(args.in_dir, args.out_dir)
    sys.exit(return_code)


# ---------------------------------------------------------------------------
# Run 'main' if the script is not imported as a module
# ---------------------------------------------------------------------------
# if __name__ == '__main__':
#     main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Read LCHEAPO data into an obspy stream
"""
import argparse
import re

from obspy.core import UTCDateTime, Stream

from .chan_maps import chan_maps
from .lcread import read, get_data_timelimits


def main():
    """
    Command-line plotting interface
    """
    obs_types = [s for s in chan_maps]
    epilog = "--sfilt returns the first parenthetised subgroup, using \n"
    epilog += "  python's re.search().  Some useful codes are: \n"
    epilog += "   '.': matches any character\n"
    epilog += "   '+': matches 1 or more repetitions of the preceding RE\n"
    epilog += "   '?': matches 0 or 1 repetitions of the preceding RE\n"
    epilog += "  '+?': like +, but non-greedy (don't match as many as possible)\n"
    epilog += "   '\\': escapes special characters, like '*', '-' and '/')\n"
    epilog += "  '()': delimit the subgroup to return\n\n"
    epilog += " Some examples are: \n"
    epilog += "    FILENAME                 |   PATTERN      |     RESULT\n"
    epilog += "    =========================+================+=============\n"
    epilog += "    haha-MOVA-OBS1-blah.lch | '\-(.+)\-'     | MOVA-OBS1 \n"
    epilog += "    haha-MOVA-OBS1-blah.lch | '\(-.+)\-.+\-' | MOVA \n"
    epilog += "    haha-MOVA-OBS1-blah.lch | '\-.+\-(.+)\-' | OBS1 \n"
    epilog += "    MOVA/haha.raw.lch       | '([^\/]+)'     | MOVA \n"

    parser = argparse.ArgumentParser(
        description=__doc__, epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("infiles", nargs="+", help="Input filename(s)")
    parser.add_argument("-s", "--start", dest="starttime", metavar="TIME",
                        default=0,
                        help="start time (ISO8601, or seconds after last file "
                             "start) (default: %(default)s)")
    parser.add_argument("-e", "--end",
                        dest="endtime", metavar="TIME", default=3600.,
                        help="end time (ISO8601, or seconds from start time) "
                        "(default: %(default)s)")
    parser.add_argument("-t", "--type", choices=obs_types,
                        dest="obs_type", metavar='TYPE', default='SPOBS2',
                        help="obs type.  Allowed choices are " +
                             ", ".join(obs_types) +
                             " (default: %(default)s)")
    parser.add_argument("--net", dest="network", default='NN',
                        help="network code (default: %(default)s)")
    parser.add_argument("--chan", dest="channel", default='*',
                        help="Plot only the given SEED channel/s (default: "
                             "%(default)s)")
    parser.add_argument("--equal_scale", action="store_true",
                        help="Force all traces equal y scale")
    parser.add_argument('-v', "--verbose",
                        dest="verbose", action="store_true",
                        help="Print information about the first and last "
                             "read blocks")
    my_group = parser.add_mutually_exclusive_group(required=False)
    my_group.add_argument("--sta", dest="station", default='STA',
                          help="station code.  A 2-digit counter will be "
                               "appended if more than one file is read. "
                               "(default: %(default)s)")
    my_group.add_argument("--sfilt", dest="station_filt",
                          help="regex filter to find station name in filename "
                               "(see examples below)")
    args = parser.parse_args()

    # Set/normalize start and end times
    endtime = _normalize_time_arg(args.endtime)
    if endtime == 0:
        endtime = 3600.
    if not args.starttime:  # set starttime to latest-starting file
        args.starttime = UTCDateTime(0)
        for infile in args.infiles:
            s, e = get_data_timelimits(infile)
            if s > args.starttime:
                args.starttime = s
    # Read file(s)
    station_code = None
    if args.station:
        if len(args.infiles) == 1:
            station_code = args.station
    stream = Stream()
    for (infile, i) in zip(args.infiles, range(len(args.infiles))):
        if args.station_filt:
            # print(re.search(args.station_filt, infile))
            try:
                station_code = re.search(args.station_filt, infile).group(1)
            except Exception:
                print('no station code found using re.search("{}", "{}"'.
                      format(args.station_filt, infile))
                station_code = None
        if station_code is None:
            station_code = f'STA{i:02d}'
        s = read(infile,
                 _normalize_time_arg(args.starttime),
                 _normalize_time_arg(args.endtime),
                 network=args.network,
                 station=station_code,
                 obs_type=args.obs_type,
                 verbose=args.verbose)
        if s is not None:
            s = s.select(channel=args.channel)
            stream += s
            station_code = None
    if len(stream) > 0:
        stream.plot(size=(800, 600), equal_scale=args.equal_scale,
                    method='full')
    else:
        print('Nothing read, nothing plotted!')


def _normalize_time_arg(a):
    """
    Convert time from string to float if it is numeric
    """
    if isinstance(a, UTCDateTime):
        return a
    try:
        temp = float(a)
    except ValueError:
        return a
    else:
        return temp

# ---------------------------------------------------------------------------
# Run 'main' if the script is not imported as a module
# ---------------------------------------------------------------------------
# if __name__ == '__main__':
#     main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Read LCHEAPO data into an obspy stream
"""
import warnings
import struct
import os
import sys
import inspect

import numpy as np
from obspy.core import UTCDateTime, Stream, Trace
from obspy import read_inventory

from .lcheapo_utils import (LCDataBlock, LCDiskHeader)
from .chan_maps import chan_maps


def read(filename, starttime=None, endtime=None, network='XX', station='SSSSS',
         obs_type=None, verbose=False):
    """
    Read LCHEAPO data into an obspy stream

    To avoid overlaps, starttime is inclusive and endtime is exclusive

    Args:
        filename (str): LCHEAPO filename
        starttime (:class:`~obspy.core.utcdatetime.UTCDateTime`): 
            Start time as a ISO8601 string, a UTCDateTime object,
            or a number (seconds after the file start)
        endtime (:class:`~obspy.core.utcdatetime.UTCDateTime`):
            End time as a ISO8601 string, a UTCDateTime object,
            or a number (seconds after starttime)
        network (str): Set network code (up to two characters)
        station (str): Set FDSN station name (up to five characters)
        obs_type (str): OBS type (must match a key in chan_maps)
        verbose (bool): print out info about first and last read data

    Returns: 
        stream (:class:`~obspy.core.stream.Stream`): read data

    Example:
        >>> from lcheapo import read
        >>> st = read("/path/to/four_channels.lch")
        >>> print(st)  # doctest: +ELLIPSIS
        2 Trace(s) in Stream:
        BW.UH3..EHE | 2010-06-20T00:00:00.279999Z - ... | 200.0 Hz, 386 samples
        BW.UH3..EHZ | 2010-06-20T00:00:00.279999Z - ... | 200.0 Hz, 386 samples
    """
    # Check/complete input variables
    if len(network) > 2:
        network = network[:2]
    if len(station) > 5:
        network = network[:5]
    if not obs_type:
        warnings.warn('No obs_type provided, assuming SPOBS2')
        obs_type = 'SPOBS2'

    with open(filename, 'rb') as fp:
        data = _read_data(starttime, endtime, fp, verbose)
        if data is None:
            print(f'Did not read from file {filename}')
            return None
    data = _stuff_info(data, network, station, obs_type)
    return data


def get_data_timelimits(lcheapo_object):
    """
    Return data start and end times

    Args:
        lcheapo_object (str or file-like object):
            filename or open file-like object that contains the
            binary Mini-SEED data.  Any object that provides a read()
            method will be considered a file-like object.
    Returns:
        tuple (tuple): 2-tuple:
            startime (:class:`obspy.core.UTCDateTime``): start of data
            endime (:class:`obspy.core.UTCDateTime``): end of data
    """

    if hasattr(lcheapo_object, "read"):
        fp = lcheapo_object
    else:
        fp = open(lcheapo_object, 'rb')

    lcHeader = LCDiskHeader()
    block = LCDataBlock()
    status = lcHeader.readHeader(fp)
    if status == 0:
        return None, None

    # read starttime
    startBlock = lcHeader.dataStart
    block.seekBlock(fp, startBlock)
    block.readBlock(fp)
    starttime = block.getDateTime()

    # read endtime
    lastBlock = block.determineLastBlock(fp)
    block.seekBlock(fp, lastBlock)
    block.readBlock(fp)
    endtime = block.getDateTime()

    return UTCDateTime(starttime), UTCDateTime(endtime)


def band_code_sps(band_code, sps):
    """
    Verify/correct a channel's band code for a given sampling rate.
    Band codes are from
    http://docs.fdsn.org/projects/source-identifiers/en/v1.0/channel-codes.html

    Args:
        band_code (str): SEED channel code
        sps (float): sampling rate

    Returns:
        new_band_code (str): SEED channel band code corresponding to
            sampling rate
    """
    # print(f"{band_code=}, {sps=}")
    band_codes_SP = 'GDES'
    band_codes_LP = 'FCHBMLVUWRPTQ'
    if sps > 5000:
        return 'J'
    if band_code in band_codes_SP:
        if sps >= 1000:
            return "G"
        elif sps >= 250:
            return "D"
        elif sps >= 80:
            return "E"
        elif sps >= 10:
            return "S"
        else:
            raise NameError(
                "You should not have short period data at < 10 sps")
    elif band_code in band_codes_LP:
        if sps >= 1000:
            return "F"
        elif sps >= 250:
            return "C"
        elif sps >= 80:
            return "H"
        elif sps >= 10:
            return "B"
        elif sps > 1:
            return "M"
        elif sps == 1:
            return "L"
        elif sps >= 0.1:
            return "V"
        elif sps >= 0.01:
            return "U"
        elif sps >= 0.001:
            return "W"
        elif sps >= 0.0001:
            return "R"
        elif sps >= 0.00001:
            return "P"
        elif sps >= 0.000001:
            return "T"
        else:
            return "Q"
    raise NameError(f'Unknown band code "{band_code}"')


def _read_data(starttime, endtime, fp, verbose=False):
    """
    Return data.

    Returns from the start of the block containing starttime to the end of the
    block containing endtime

    Args:
        starttime (:class:`~obspy.UTCDateTime`): start time
        endtime (:class:`~obspy.UTCDateTime`): end time
        fp (:class:`file`): file pointer
    
    Returns
        stream (:class:`obspy.core.Stream`):

    For speed, reads all blocks at once and extracts as slices
    """
    starttime, endtime = _convert_time_bounds(starttime, endtime, fp)
    if starttime is None:
        return None

    lcHeader = LCDiskHeader()
    block = LCDataBlock()
    lcHeader.readHeader(fp)
    # lcHeader.printHeader()
    sample_rate = lcHeader.realSampleRate
    n_chans = lcHeader.numberOfChannels
    n_start_block = _get_block_number(starttime, fp)
    n_end_block = _get_block_number(endtime, fp) + n_chans - 1
    stream = Stream()

    chan_blocks = int(((n_end_block - n_start_block + 1) / n_chans))
    read_blocks = chan_blocks * n_chans
    block.seekBlock(fp, n_start_block)

    # Read the data and arrange in read_blocks*512 array
    buf = fp.read(read_blocks * 512)
    dt = np.dtype('b')
    all = np.frombuffer(buf, dtype=dt)
    a = np.reshape(all, (read_blocks, 512))  # keep data contiguous
    # The following takes the same amount of time ...
    # a = all.view()
    # a.shape = (read_blocks,512)
    headers = a[:, :14]
    data = a[:, 14:]

    # Get header information and determine if data are contiguous
    samples_per_block = _get_header_nsamples(headers[0, :])
    seconds_per_block = samples_per_block / sample_rate
    last_time = _get_header_time(headers[-n_chans, :])
    first_time = _get_header_time(headers[0, :])
    if verbose:
        print(f'First read block = {n_start_block:d}, time = {first_time}')
        print(f'Last read block =  {n_end_block:d}, time = {last_time}')
    timerange = last_time - first_time
    expected_timerange = (chan_blocks - 1) * seconds_per_block
    offset = timerange - expected_timerange
    if offset > 0.1/sample_rate:
        warnings.warn(f"Last block late by {offset:g} seconds! "
                      f"({offset*sample_rate:g} samples, "
                      f"{offset/seconds_per_block:g} blocks)")
    elif offset < -0.1/sample_rate:
        warnings.warn(f"Last block early by {-offset:g} seconds! "
                      f"({-offset*sample_rate:g} samples, "
                      f"{-offset/seconds_per_block:g} blocks)")
    stats = {'sampling_rate': sample_rate, 'starttime': first_time}

    # Extract channels
    for i in range(0, n_chans):
        # Get data
        chan_data = data[i:read_blocks:n_chans, :].flatten()
        # could be quicker using the np.flat() iterator ?
        t32 = np.array(chan_data[0: 498*chan_blocks: 3]*(1 << 16)
                       + chan_data[1: 498*chan_blocks: 3].astype('B')
                       * (1 << 8)
                       + chan_data[2: 498*chan_blocks: 3].astype('B'),
                       dtype='int32')
        stream.append(Trace(data=t32, header=stats))
    eps = 1e-6
    stream.trim(starttime=starttime, endtime=endtime-eps, nearest_sample=False)
    return stream


def _get_header_time(header):
    """
    Return time from an LCHEAPO header

    Args:
        header: 14-byte LCHEAPO HEADER
    """
    (msec, second, minute, hour, day, month, year) = struct.unpack(
        '>HBBBBBB', header.data[:8])
    try:
        return UTCDateTime(year + 2000, month, day, hour, minute,
                           second + msec/1000)
    except Exception as err:
        raise ValueError(
            '{}. header fields= {}, {}, {}, {}, {}, {}, {}'
            .format(err, year, month, day, hour, minute, second, msec))


def _get_header_nsamples(header):
    """
    Return number of samples from an LCHEAPO header

    Args:
        header = 14-byte LCHEAPO HEADER
    """
    (U1, U2) = struct.unpack('>BB', header.data[12:])
    return U2


def _convert_time_bounds(starttime, endtime, fp):
    """
    Return starttime and endtime as UTCDateTimes

    Also makes sure that starttime and endtime are within the data bounds
    
    Args:
        startime (str, float or UTCDate time): starttime.  If float, 
            seconds from file start
        endtime (str, float or UTCDate time): end time.  If float, seconds
            after starttime
        fp: file pointer (for checking if the times are within the
            data bounds)
    """
    data_start, data_end = get_data_timelimits(fp)
    if data_start is None:
        return None, None
    if not starttime:
        starttime = 0
    if endtime == 0:
        endtime = data_end
    if not endtime:
        endtime = 3600
    if isinstance(starttime, (float, int)):
        starttime = data_start + starttime
    else:
        starttime = UTCDateTime(starttime)
    if isinstance(endtime, (float, int)):
        endtime = starttime + endtime
    # Handle bad time ranges
    if endtime <= starttime:
        print("{endtime=} is before {starttime=}")
        return None, None
    if starttime >= data_end:
        print(f"{starttime=} is after {data_end}=")
        return None, None
    if endtime <= data_start:
        print(f"{endtime=} is before {data_start=}")
        return None, None
    if starttime < data_start:
        starttime = data_start
    if endtime > data_end:
        endtime = data_end
    return UTCDateTime(starttime), UTCDateTime(endtime)


def _get_block_number(time, fp):
    """
    Return block number containing first channel containing the given time

    Args:
        time (:class:`~obspy.UTCDateTime`): the time
        fp (:class:`~file`): file pointer
    """
    lcHeader = LCDiskHeader()
    block = LCDataBlock()
    lcHeader.readHeader(fp)
    data_start_block = lcHeader.dataStart
    block.seekBlock(fp, data_start_block)
    block.readBlock(fp)

    starttime, _ = get_data_timelimits(fp)
    block_len_s = block.numberOfSamples / lcHeader.realSampleRate
    num_chans = lcHeader.numberOfChannels
    record_offset = int((UTCDateTime(time)-starttime) / block_len_s)

    return data_start_block + record_offset * num_chans


def _stuff_info(stream, network, station, obs_type):
    """
    Put network, station and channel information into station stream

    Args:
        stream (:class:`~obspy.stream.Stream`): obspy stream
        network (str): network code
        station (str): station code
        obs_type (str): type of obs (must be in channel_maps)

    Returns:
        data (:class:`~obspy.stream.Stream`): informed data
    """
    assert obs_type in chan_maps
    chan_map = chan_maps[obs_type]
    # chan_map_list = list(chan_map)
    for trace, chan_loc in zip(stream, chan_map):
        trace.stats.network = network
        trace.stats.station = station
        sps = trace.stats.sampling_rate
        # if len(chan_map) >= 1:
        chan, loc = chan_loc.split(':')
        trace.stats.channel = band_code_sps(chan[0], sps) + chan[1:3]
        if len(loc) > 1:
            trace.stats.location = loc
        trace.stats.response = _load_response(obs_type, trace.stats.channel,
                                              trace.stats.starttime)
    return stream


def _load_response(obs_type, channel, start_time):
    """
    Load response corresponding to OBS type and component

    Args:
        obs_type (str): obs type (must be in channel_maps)
        channel (str): trace channel code
        datetime (:class:`~obspy.UTCDateTime`): time for which to get response

    Returns:
        resp (:class:`~obspy.core.response.Response`): instrument response
    """
    assert obs_type in chan_maps
    try:
        basepath = os.path.dirname(os.path.abspath(inspect.getfile(
                                       inspect.currentframe())))
        inv_file = os.path.join(basepath, 'data', f'{obs_type}.station.xml')
        inv = read_inventory(inv_file)
    except Exception:
        print(f'Could not read inventory file {inv_file}')
        sys.exit()

    try:
        resp = inv.select(channel=channel, time=start_time)[0][0][0].response
    except Exception:
        print(f'No response matching "{channel}" at {start_time}')
        print('Options were: ')
        for net in inv:
            for sta in net:
                for ch in sta:
                    print(' "{}" : {} - {}'.format(ch.code, ch.start_date,
                                                   ch.end_date))
        return []
    return resp


def _valid_chan_map(chan_map):
    """
    Verify that chan_map is valid
    """
    if not isinstance(chan_map, list):
        return False
    for elem in chan_map:
        if not isinstance(elem, str):
            return False
        if len(elem) <= 3:
            return True
        else:
            if ':' in elem:
                elems = elem.split(':')
                if len(elems) == 2:
                    if len(elems[0]) <= 3 and len(elems[1]) <= 2:
                        return True
    return False

### THE FOLLOWING DUPICATES LCPLOT.PY BECAUSE I CAN'T CHANGE MODULE REFERENCES
### WITHOUT INTERNET


import argparse
import re

def _plot_command():
    """
    Command-line plotting interface
    """
    obs_types = [s for s in chan_maps]
    epilog = "--sfilt returns the first parenthetised subgroup, using \n"
    epilog += "  python's re.search().  Some useful codes are: \n"
    epilog += "   '.': matches any character\n"
    epilog += "   '+': matches 1 or more repetitions of the preceding RE\n"
    epilog += "   '?': matches 0 or 1 repetitions of the preceding RE\n"
    epilog += "  '+?': like +, but non-greedy (don't match as many as possible)\n"
    epilog += "   '\\': escapes special characters, like '*', '-' and '/')\n"
    epilog += "  '()': delimit the subgroup to return\n\n"
    epilog += " Some examples are: \n"
    epilog += "    FILENAME                 |   PATTERN      |     RESULT\n"
    epilog += "    =========================+================+=============\n"
    epilog += "    haha-MOVA-OBS1-blah.lch | '\-(.+)\-'     | MOVA-OBS1 \n"
    epilog += "    haha-MOVA-OBS1-blah.lch | '\(-.+)\-.+\-' | MOVA \n"
    epilog += "    haha-MOVA-OBS1-blah.lch | '\-.+\-(.+)\-' | OBS1 \n"
    epilog += "    MOVA/haha.raw.lch       | '([^\/]+)'     | MOVA \n"

    parser = argparse.ArgumentParser(
        description=__doc__, epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("infiles", nargs="+", help="Input filename(s)")
    parser.add_argument("-s", "--start", dest="starttime", metavar="TIME",
                        default=0,
                        help="start time (ISO8601, or seconds after last file "
                             "start) (default: %(default)s)")
    parser.add_argument("-e", "--end",
                        dest="endtime", metavar="TIME", default=3600.,
                        help="end time (ISO8601, or seconds from start time) "
                        "(default: %(default)s)")
    parser.add_argument("-t", "--type", choices=obs_types,
                        dest="obs_type", metavar='TYPE', default='SPOBS2',
                        help="obs type.  Allowed choices are " +
                             ", ".join(obs_types) +
                             " (default: %(default)s)")
    parser.add_argument("--net", dest="network", default='NN',
                        help="network code (default: %(default)s)")
    parser.add_argument("--chan", dest="channel", default='*',
                        help="Plot only the given SEED channel/s (default: "
                             "%(default)s)")
    parser.add_argument("--equal_scale", action="store_true",
                        help="Force all traces equal y scale")
    parser.add_argument('-v', "--verbose",
                        dest="verbose", action="store_true",
                        help="Print information about the first and last "
                             "read blocks")
    my_group = parser.add_mutually_exclusive_group(required=False)
    my_group.add_argument("--sta", dest="station", default='STA',
                          help="station code.  A 2-digit counter will be "
                               "appended if more than one file is read. "
                               "(default: %(default)s)")
    my_group.add_argument("--sfilt", dest="station_filt",
                          help="regex filter to find station name in filename "
                               "(see examples below)")
    args = parser.parse_args()

    # Set/normalize start and end times
    endtime = _normalize_time_arg(args.endtime)
    if endtime == 0:
        endtime = 3600.
    if not args.starttime:  # set starttime to latest-starting file
        args.starttime = UTCDateTime(0)
        for infile in args.infiles:
            s, e = get_data_timelimits(infile)
            if s > args.starttime:
                args.starttime = s
    # Read file(s)
    station_code = None
    if args.station:
        if len(args.infiles) == 1:
            station_code = args.station
    stream = Stream()
    for (infile, i) in zip(args.infiles, range(len(args.infiles))):
        if args.station_filt:
            # print(re.search(args.station_filt, infile))
            try:
                station_code = re.search(args.station_filt, infile).group(1)
            except Exception:
                print('no station code found using re.search("{}", "{}"'.
                      format(args.station_filt, infile))
                station_code = None
        if station_code is None:
            station_code = f'STA{i:02d}'
        s = read(infile,
                 _normalize_time_arg(args.starttime),
                 _normalize_time_arg(args.endtime),
                 network=args.network,
                 station=station_code,
                 obs_type=args.obs_type,
                 verbose=args.verbose)
        if s is not None:
            s = s.select(channel=args.channel)
            stream += s
            station_code = None
    if len(stream) > 0:
        stream.plot(size=(800, 600), equal_scale=args.equal_scale,
                    method='full')
    else:
        print('Nothing read, nothing plotted!')


def _normalize_time_arg(a):
    """
    Convert time from string to float if it is numeric
    """
    if isinstance(a, UTCDateTime):
        return a
    try:
        temp = float(a)
    except ValueError:
        return a
    else:
        return temp


# ---------------------------------------------------------------------------
# Run 'main' if the script is not imported as a module
# ---------------------------------------------------------------------------
# if __name__ == '__main__':
#     main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
SDPCHAIN argparse wrapper
"""
from os import mkdir
from pathlib import Path
import argparse
import sys

# from .version import __version__


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, add_infile=False, add_infiles=False, add_outfile=False,
                 add_outfiles=False, force_outfile=False, **kwargs):
        """
        Create argparse instance with -d, -i, -o options already specified

        infile(s) are added without
        outfile is added as --of
        outfiles is added as --ofs
        force_outfile (-F) overwrites the outfile if it exists already

        for in_path, out_path, in_file, in_files, out_file, out_files, returns
        the entered values in attributes with the above names, and the fully
        expanded values in full_{NAME} for each of the above names

        Also validates that the infile(s) exist and that the outfile(s) do not
        """
        assert not (add_infile and add_infiles)
        assert not (add_outfile and add_outfiles)

        super(ArgumentParser, self).__init__(**kwargs)

        if add_infile:
            self.add_argument("infile", help="Input filename")
        elif add_infiles:
            self.add_argument("infiles", help="Input filenames")

        if add_outfile:
            self.add_argument("--of", dest="outfile", required=True,
                              help="Output filename")
        elif add_outfiles:
            self.add_argument("--ofs", dest="outfiles", required=True,
                              help="Output filenames")

        self.add_argument("-d", "--directory", dest="base_dir",
                          default='.', help="Base directory for files")
        self.add_argument("-i", "--input", dest="in_dir", default='.',
                          help="path to input files (abs, or rel to base)")
        self.add_argument("-o", "--output", dest="out_dir", default='.',
                          help="path for output files (abs, or rel to base)")

        if force_outfile:
            self.add_argument("-F", "--force_outfile", action="store_true",
                              help="Force overwrite of output files")

    def parse_args(self, **kwargs):
        """
        Parse arguments, processing in_dir and out_dir relative to base_dir
        
        Also sets up input and output file paths, verifies the existence of
        input files, the non-existence of output files, and copies any
        process-steps.json file in the input directory to the base directory
        (should we also copy to the output directory?)
        """
        args = self.parse_args(**kwargs)
        full_in, full_out = setup_paths(args)
        # Save full paths to "full_*"
        args.setattr('full_in_dir', full_in)
        args.setattr('full_out_dir', full_out)
        
        # if there is a process-steps file in the input directory, copy it
        # to the base directory
        if args.in_dir:
            if (Path(full_in_dir) / "process-steps.json").is_file() and
                    not (Path(base_dir) / "process-steps.json").exists():
                shutil.copy2(Path(full_in_dir) / "process-steps.json",
                             Path(args.base_dir) / "process-steps.json")

        # Create full path versions of input files and verify their existence
        if getattr(args, "infile", None) is not None:
            args.setattr("full_infile", Path(full_in) / args.infile)
            if not args.full_infile.is_file():
                print('input file {args.infile} does not exist, quitting')
                sys.exit(2)
        elif getattr(args, "infiles", None) is not None:
            args.setattr("full_infiles",
                         [Path(full_in) / f for f in args.infiles])
            for f in args.full_infiles:
                if not f.is_file():
                    print('input file {f.name} does not exist, quitting')
                    sys.exit(2)

        # Create full path versions of output files and verify non-existence
        if getattr(args, 'force_outfile', False) is True:
            if getattr(args, "outfile", None) is not None:
                args.setattr("full_outfile", Path(full_out) / args.outfile)
                if args.full_outfile.exists():
                    print('output file {args.outfile} exists, quitting')
                    sys.exit(2)
            elif getattr(args, "outfiles", None) is not None:
                args.setattr("full_outfiles",
                             [Path(full_out) / f for f in args.outfiles])
                for of in args.full_outfiles:
                    if of.exists():
                        print('output file {of.name} exists, quitting')
                        sys.exit(2)

        return args


def setup_paths(base_dir, in_dir, out_dir):
    """
    Set up paths using SDPCHAIN standards

    :parm base_dir: base directory (for process-steps.json file and as
                    a basis for in_dir and out_dir)
    :param in_dir: directory for input files.  absolute path or relative
                   to base_dir
    :param out_dir: directory for ourput files.  absolute path or relative
                   to base_dir
                 - out_dir directory containing output files
    :return in_dir, out_dir: base_dir-adjusted paths
    """
    in_path = _choose_path(base_dir, in_dir)
    out_path = _choose_path(base_dir, out_dir)
    assert Path(in_path).is_dir()
    if Path(out_path).exists() is False:
        print(f"out_dir '{out_path}' does not exist, creating...")
        mkdir(out_path)
    elif Path(out_path).is_file():
        print("out_dir '{out_path}' is a file! Will use  base dir")
        out_path = base_dir
    return in_path, out_path


def _choose_path(base_dir, sub_dir):
    """ Sets up absolute path to sub-directory """
    if Path(sub_dir).is_absolute():
        return sub_dir
    return str(Path(base_dir) / sub_dir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Read LCHEAPO data into an obspy stream
"""
import argparse
import warnings
# import os
import sys
import datetime
import inspect
from pathlib import Path

from obspy.core import UTCDateTime
from .sdpchain import ProcessStep
from progress.bar import IncrementalBar

from .chan_maps import chan_maps
from .lcread import read as lcread, get_data_timelimits
from .version import __version__


def lc2SDS():
    """
    Convert fixed LCHEAPO data to SeisComp Data Structure

    SIMPLE drift and leapsecond correction:
        - offset is constant within each daily file
        - offset information is not written in header
        - data quality field is not modified
        - leapsecond flag is not raised (causes apparent 1-s gap/overlap).
    Doesn't combine one day's SDS files made from two different LCHEAPO files
      (last one overwrites first)
    Writes to a directory named SDS/ in the output directory.
    """
    # print(lc2SDS.__doc__)
    parser = argparse.ArgumentParser(
        description=inspect.cleandoc(lc2SDS.__doc__),
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("input_files", nargs='+',
                        help="Input filename(s).  If there are captured "
                             "wildcards (put in '' so that they aren't "
                             "interpreted by the shell), will expand them "
                             "in the input directory")
    parser.add_argument("-t", "--obs_type", default='SPOBS2',
                        help="obs type.  Controls channel and location codes",
                        choices=[s for s in chan_maps])
    parser.add_argument("--station", default='SSSSS',
                        help="station code for this instrument")
    parser.add_argument("--network", default='XX',
                        help="network code for this instrument")
    parser.add_argument("-s", "--start_times", nargs='+',
                        metavar=("REF_START", "INST_START"),
                        help="Start datetimes for the reference (usually GPS) "
                             "and instrument.  If only one value is provided, "
                             "it will be used for both")
    parser.add_argument("-e", "--end_times", nargs=2,
                        metavar=("REF_END", "INST_END"),
                        help="End datetimes for the reference and instrument")
    parser.add_argument("--leapsecond_times", nargs='+',
                        help="leapsecond times")
    parser.add_argument("--leapsecond_types", default='+',
                        help="'+' for extra second, '-' for removed second. "
                             "If there is one character it is applied to all "
                             "leapseconds, if there is more than one the "
                             "length of the string must match "
                             "the number of leapsecond_times")
    parser.add_argument("-d", dest="base_dir", metavar="BASE_DIR",
                        default='.', help="base directory for files")
    parser.add_argument("-i", dest="in_dir", metavar="IN_DIR", default='.',
                        help="input file directory (absolute, " +
                             "or relative to base_dir)")
    parser.add_argument("-o", dest="out_dir", metavar="OUT_DIR", default='.',
                        help="output file directory (absolute, " +
                             "or relative to base_dir)")
    parser.add_argument("-v", "--verbose", action='store_true',
                        help="verbose output")
    parser.add_argument("--version", action='store_true',
                        help="Print version number and quit")
    args = parser.parse_args()
    parameters = vars(args).copy()
    if args.version is True:
        print(f"Version {__version__}")
        sys.exit(0)
    
    # ADJUST INPUT PARAMETERS
    if args.start_times is not None:
        args.start_times = [UTCDateTime(x) for x in args.start_times]
    if args.end_times is not None:
        args.end_times = [UTCDateTime(x) for x in args.end_times]
    ls_times, ls_types = _adjust_leapseconds(args.leapsecond_times,
                                             args.leapsecond_types)

    # SETUP FOR PROCESS-STEPS
    process_step = ProcessStep('lc2SDS',
                               " ".join(sys.argv),
                               app_description=__doc__,
                               app_version=__version__,
                               parameters=parameters)
    args.in_dir, args.out_dir, args.input_files = ProcessStep.setup_paths(args)
    # Expand captured wildcards
    #args.input_files = [x.name for f in args.input_files
    #                for x in Path(args.in_dir).glob(f)]

    for infile in args.input_files:
        lc_start, lc_end = get_data_timelimits(Path(args.in_dir) / infile)

        if args.start_times and args.end_times:
            ref_start = args.start_times[0]
            if len(args.start_times) > 1:
                inst_start = args.start_times[1]
            else:
                inst_start = ref_start
            ref_end, inst_end = args.end_times
            if inst_start == 0:
                inst_start = ref_start
            inst_start_offset = inst_start - ref_start
            inst_drift = ((inst_end - ref_end) - inst_start_offset)\
                / (ref_end - inst_start)
            print('instrument start offset = {:g}s, drift rate = {:.4g}'
                  .format(inst_start_offset, inst_drift))
            # quality_flag = 'Q'  # Don't know how to put this in miniSEED
        else:
            ref_start, inst_start = lc_start, lc_start
            inst_start_offset = 0
            inst_drift = 0
            warnings.warn('Could not calculate clock drift, assuming zero!')
            # quality_flag = 'D'  # Don't know how to put this in miniSEED

        lc_start_day = lc_start.replace(hour=0, minute=0, second=0,
                                        microsecond=0)
        lc_end_day = lc_end.replace(hour=0, minute=0, second=0, microsecond=0)
        stime = lc_start_day
        bar = IncrementalBar(f'Processing {infile}',
                             max=(lc_end_day-lc_start_day)/86400 + 1)
        while stime <= lc_end_day:
            inst_offset = inst_start_offset + inst_drift * (stime - ref_start)
            _write_daily(inst_offset, stime, infile, args, ls_times, ls_types)
            bar.next()
            stime += 86400
        bar.finish()

    return_code = 0
    process_step.exit_code = return_code
    process_step.write(args.in_dir, args.out_dir)
    sys.exit(return_code)


def _write_daily(inst_offset, stime, infile, args, ls_times, ls_types):
    starttime = stime + inst_offset
    endtime = starttime + 86400
    if args.verbose:
        print('{}, inst_offset = {:.3f}s: reading {}-{}'.format(
            stime.strftime('%Y-%m-%d'), inst_offset,
            starttime.isoformat(), endtime.isoformat()))
    stream = lcread(Path(args.in_dir) / infile,
                    starttime=starttime,
                    endtime=endtime,
                    network=args.network,
                    station=args.station,
                    obs_type=args.obs_type)

    for tr in stream:
        s = tr.stats
        # Correct leapsecond
        s.starttime += _leap_correct(stime, ls_times, ls_types)
        # Correct drift
        s.starttime -= inst_offset
        # s.mseed['dataquality'] = quality_flag

        # Write file
        dirname = Path(args.out_dir) / 'SDS' / str(stime.year) /\
            s.network / s.station / f'{s.channel}.D'
        fname = '{}.{}.{}.{}.D.{}.{:03d}'.format(
            s.network, s.station, s.location, s.channel,
            stime.year, stime.julday)
        dirname.mkdir(parents=True, exist_ok=True)
        tr.write(str(dirname / fname), format='MSEED',
                 encoding='STEIM1', reclen=4096)


def _adjust_leapseconds(ls_times, ls_types):
    """
    Adjust leapsecond arguments
    
    Converts times to UTCDateTime and makes sure there is one type
    for each time
    :param ls_times: list of strings
    :param ls_types: str
    :returns: ls_times, ls_types
    """
    if ls_times is None:
        return ls_times, ls_types
    # Quick and stupid way to avoid second=60 problem
    new_times = []
    for t in ls_times:
        if t[-2:] == '60':
            t=t[:-2] + '59'
        new_times.append(UTCDateTime(t))
    if not len(new_times) == len(ls_types):
        if len(ls_types) == 1:
            ls_types = ls_types * len(new_times)
        else:
            raise IndexError(f"len(ls_times) ({len(new_times)}) "
                             "incompatible with "
                             f"len(ls_types) {len(ls_types)}")
    for tp in ls_types:
        if tp not in '+-':
            raise ValueError(f"'{tp}' is not a valid leapsecond type")
    return new_times, ls_types

                            
def _leap_correct(starttime, ls_times, ls_types):
    """
    Return leap-second correction for a given time
    :param starttime: time at start of current data segment
    :type starttime: UTCDateTime
    :param ls_times: list of leap second times.
    :param ls_types: str of leap second types ('+' or '-').  If len(ls_times)
        is not 0, ls_types and ls_times must have same length
    :returns: seconds to add to current OBS-recorded time
    """
    if ls_times is None:
        return 0
    correct = 0
    for ls_time, ls_type in zip(ls_times, ls_types):
        if starttime > ls_time - 1: # avoid overlap with leap seconds
            if ls_type=='+':
                correct -= 1  # Everything after an added second is one earlier
            elif ls_type=='-':
                correct += 1
            else:
                raise ValueError(f"'{ls_type}' is not a valid leapsecond type")
    return correct
        
        
# ---------------------------------------------------------------------------
# Run 'main' if the script is not imported as a module
# ---------------------------------------------------------------------------
# if __name__ == '__main__':
#     main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Concatenate files and create/append process-step.json file

Uses python standard library's shutil.copyfileobj()
"""
# from __future__ import (absolute_import, division, print_function,
#                         unicode_literals)
# from future.builtins import *  # NOQA @UnusedWildImport

import argparse
# import os
from datetime import datetime as dt
import sys
from pathlib import Path
import shutil

from .process_steps import ProcessStep
from .version import __version__


def sdpcat():
    # GET ARGUMENTS
    args = getOptions()
    process_step = ProcessStep(
        'sdpcat',
        " ".join(sys.argv),
        app_description = __doc__,
        app_version = __version__,
        parameters=args)

    args.in_dir, args.out_dir, _ = ProcessStep.setup_paths(args)
    # VERIFY THAT INPUT FILES EXIST AND OUTPUT FILE DOESN'T
    if len(args.input_files) < 2:
        print("You only provided 1 input file: this is just a copy!")
    for file in args.input_files:
        f = Path(args.in_dir) / file
        assert Path(f).is_file()
    assert (Path(args.out_dir) / args.output_file).exists() is False

    with open(Path(args.out_dir) / args.output_file, mode="wb") as destination:
        for file in args.input_files:
            with open(Path(args.in_dir) / file, mode="rb") as source:
                shutil.copyfileobj(source, destination)

    # Save/append process information to process_steps file
    process_step.exit_status = 0
    process_step.write(args.in_dir, args.out_dir)


def getOptions():
    """
    Parse user passed options and parameters.
    """
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("--of", dest="output_file", required=True, help="Output filename")
    parser.add_argument("--ifs", dest="input_files", required=True, nargs='+',
                        help="Input filenames")
    parser.add_argument("-d", "--directory", dest="base_dir",
                        default='.', help="Base directory for files")
    parser.add_argument("-i", "--input", dest="in_dir", default='.',
                        help="path to input files (absolute, or relative to "
                             "base)")
    parser.add_argument("-o", "--output", dest="out_dir", default='.',
                        help="path for output files (absolute, or relative to "
                             "base)")
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    sdpcat()



"""
Functions to calculate spectra, coherences and transfer functions
"""
from obspy.signal.invsim import cosine_taper
from obspy.core.trace import Trace
import scipy.signal as ssig
from matplotlib import mlab
import matplotlib.pyplot as plt
import numpy as np
import math as m
import warnings

from .Peterson_noise_model import PetersonNoiseModel

# Set variables
spect_library = 'scipy'  # 'mlab' or 'scipy': mlab gives weird coherences!


class PSD:
    """
    Power Spectral Density class
    """
    def __init__(self, freqs=None, data=None, units=None, stats=None):
        """
        :parm freqs: 1-D array of frequencies
        :parm data: 1-D array of PSD values
        :parm units: data units (str)
        :type stats: :class:`~obspy.core.trace.Stats`
        """
        self.freqs = np.array(freqs)
        self.data = np.array(data)
        assert freqs.shape == data.shape
        self.units = units
        self.stats = stats

    def __repr__(self):
        """
        String describing object
        """
        s = 'PSD(freqs, data, units={}, stats=stats) '.format(
            self.units, self.stats)
        s += '<data.shape={}, freqs={:g}-{:g}, seed_code={}>'.format(
            self.data.shape, self.freqs[0], self.freqs[-1],
            _seed_code(self.stats))
        return s

    @classmethod
    def calc(cls, tr, window_length=1000):
        """
        Calculate PSD of a data trace
        Based on obspy PPSD function

        REMOVING PART COHERENT WITH ANOTHER CHANNEL WILL REQUIRE DOING
        WELCH MYSELF (MUST APPLY REMOVAL AT THE LEVEL OF EACH FFT)

        :type st: :class:`~obspy.core.stream.Stream`
        :param tr: Trace to be processed, should have response attached
        :type window_length: `numeric`
        :param window_length: minimum FFT window length in seconds
        """
        assert type(tr) == Trace
        
        sampling_rate = tr.stats.sampling_rate
        data_len = tr.stats.endtime - tr.stats.starttime
        if window_length > data_len/2:
            window_length = int(data_len/3)
        nfft = 2**(m.ceil(m.log2(window_length * sampling_rate)))
        nlap = int(0.75 * nfft)

        if spect_library == 'mlab':
            spec, _freq = mlab.psd(tr.data, nfft, sampling_rate,
                                   detrend=mlab.detrend_linear,
                                   window=_fft_taper, noverlap=nlap,
                                   sides='onesided', scale_by_freq=True)
        elif spect_library == 'scipy':
            _freq, spec = ssig.welch(tr.data, sampling_rate, nperseg=nfft,
                                     detrend="linear", noverlap=nlap)
        else:
            warnings.warn('Unknown spectra library: "{}"'.format(
                spect_library))
            return False

        # leave out first entry (offset)
        spec = spec[1:]

        # Remove the response using the same conventions
        # since the power is squared we must square the sensitivity
        # determine instrument response from metadata
        try:
            resp, _ = tr.stats.response.get_evalresp_response(
                t_samp=1 / sampling_rate, nfft=nfft, output="VEL")
            resp = resp[1:]
        except Exception as e:
            msg = ("Error getting response from provided metadata:\n"
                   "%s: %s\n"
                   "Skipping time segment(s).")
            msg = msg % (e.__class__.__name__, str(e))
            warnings.warn(msg)
            resp = None

        _freq = _freq[1:]
        if resp[0]:
            # Get the amplitude response (squared)
            respamp = np.absolute(resp * np.conjugate(resp))
            # Make omega with the same conventions as spec
            w = 2.0 * m.pi * _freq
            # Remove response
            if tr.stats.response.response_stages[0].input_units.upper()[:2] == "PA":
                print('Channel {} has input_units "{}": treating as hydrophone'.
                      format(tr.stats.channel,
                             tr.stats.response.response_stages[0].input_units))
                spec = spec / respamp
                PSD_units = "dB ref 1 Pa^2/Hz"
            else:
                spec = (w**2) * spec / respamp
                PSD_units = "dB ref 1 (m/s^2)^2/Hz"
        else:
            PSD_units = "dB ref 1 count^2/Hz"
        return cls(_freq, spec, PSD_units, tr.stats)

    def plot(self, ax=None, show=True, outfile=None, show_Peterson=True):
        """
        Plot a PSD
        """
        if ax:
            pass
            # plt.gca = ax
        else:
            plt.figure()
            ax = plt.gca()
        ax.semilogx(self.freqs, 10 * np.log10(self.data))
        if show_Peterson and not _seed_code(self.stats)[-1] == 'H':
            lownoise, highnoise = PetersonNoiseModel(self.freqs, True)
            ax.semilogx(self.freqs, lownoise,'--')
            ax.semilogx(self.freqs, highnoise,'--')
        ax.set_ylabel(self.units)
        # ax.title()
        ax.set_title(_seed_code(self.stats))
        if outfile:
            plt.savefig(outfile)
        elif show:
            plt.show()


class PSDs:
    """
    List of PSD
    """
    def __init__(self, PSDs=None):
        assert isinstance(PSDs, list)
        for p in PSDs:
            assert isinstance(p,PSD)
        self.PSDs = PSDs

    def __str__(self):
        s = ''
        for PSD in self.PSDs:
            s += PSD.__str__() + '\n'

    def __len__(self):
        return len(self.PSDs)

    @classmethod
    def calc(cls, st, window_length=1000):
        """
        Calculate PSDs of a data stream

        :type st: :class:`~obspy.core.stream.Stream`
        :param st: Stream to be processed
        :type window_length: `numeric`
        :param window_length: minimum FFT window length in seconds
        :returns: list of dictionaries containing freqs, data, units, name
        """
        PSDs = []
        for tr in st:
            PSDs.append(PSD.calc(tr, window_length))
        return cls(PSDs=PSDs)

    def plot(self, outfile=None):
        """
        plot PSDs
        """
        nPSDs = len(self.PSDs)
        nRows = m.floor(m.sqrt(nPSDs))
        nCols = m.ceil(nPSDs/nRows)
        plt.figure(1)
        i = 0
        for p in self.PSDs:
            i += 1
            ax = plt.subplot(nRows, nCols, i)
            p.plot(ax=ax, show=False)
        if outfile:
            plt.savefig(outfile)
        else:
            plt.show()
        return


class Coherence:
    def __init__(self, data, chan_nums):
        self.data = data
        self.chan_nums = chan_nums


class Coherences:
    def __init__(self, freqs, cohs, num_windows, signif_level,
                 starttime, endtime, stats):
        self.freqs = np.array(freqs)
        self.cohs = cohs
        for coh in cohs:
            assert self.freqs.shape == self.cohs.shape
        self.num_window = num_windows
        self.signif_level = signif_level
        self.starttime = starttime
        self.endtime = endtime

    @classmethod
    def calc(cls, st, window_length=1000):
        """
        Calculate coherences between channels of a data stream

        TO REMOVE PART COHERENT WITH ANOTHER CHANNEL, WILL HAVE TO DO
        WELCH MYSELF (MUST APPLY REMOVAL AT THE LEVEL OF EACH FFT)

        :type st: :class:`~obspy.core.stream.Stream`
        :param tr: Stream to be processed
        :type window_length: `numeric`
        :param window_length: minimum FFT window length in seconds
        :returns: list of dictionaries containing freqs, data, units, name
        """
        cohers = []
        for i in range(len(st)-1):
            for j in range(i+1, len(st)):
                # print(i,j)
                tr_i = st[i]
                tr_j = st[j]
                tr_i.data = tr_i.data.astype(np.float64)
                tr_j.data = tr_j.data.astype(np.float64)

                # Verify that channels are compatible (same sampling rate,
                # length and starttime)
                tmp = 'Channels {:d} and {:d}'.format(i, j)
                if tr_i.stats.sampling_rate != tr_j.stats.sampling_rate:
                    warnings.warn(tmp + 'have different samp rates')
                    cohers.append([])
                    continue
                sampling_rate = tr_i.stats.sampling_rate
                if len(tr_i.data) != len(tr_j.data):
                    warnings.warn(tmp + 'have different lengths')
                    cohers.append([])
                    continue
                data_samples = len(tr_i.data)
                if abs(tr_i.stats.starttime - tr_j.stats.starttime) >\
                        1 / sampling_rate:
                    warnings.warn('tmp +  ' + 'are offset by > one sample')
                    cohers.append([])
                    continue
                starttime = tr_i.stats.starttime

                # Calculate Coherence
                nfft = 2**(m.ceil(m.log2(window_length * sampling_rate)))
                nlap = int(0.75 * nfft)
                if spect_library == 'mlab':   # MLAB GIVES STRANGE ANSWER
                    Cxy, _freq = mlab.cohere(tr_i.data, tr_j.data, nfft,
                                             sampling_rate,
                                             detrend=mlab.detrend_linear,
                                             window=_fft_taper,
                                             noverlap=nlap, sides='onesided',
                                             scale_by_freq=False)
                # SCIPY GIVES SIMILAR ANSWER TO MATLAB
                elif spect_library == 'scipy':
                    _freq, Cxy = ssig.coherence(tr_i.data, tr_j.data,
                                                sampling_rate,
                                                # window=_fft_taper,
                                                nperseg=nfft,
                                                detrend="linear",
                                                noverlap=nlap)
                else:
                    warnings.warn('Unknown spectra library: "{}"'.format(
                        spect_library))
                    return False

                nW = 1 + np.floor((data_samples - nfft) / nlap)
                csl = np.sqrt(2. / nW)  # coherency 95% significance level

                # Make dictionary, leaving out first freq/Cxy entry (offset)
                cohers.append(Coherence(Cxy[1:], (i, j)))
        return cls(_freq[1:], cohers, nW, csl, starttime,
                   starttime + data_samples / sampling_rate,
                   [s.stats for s in st])

    def plot(self, outfile=None):
        """
        plot coherences calculated using calc_cohers (should become a class)
        """
        nCohers = len(self.cohers)
        nRows = m.ceil(m.sqrt(nCohers))
        nCols = m.ceil(nCohers / nRows)
        plt.figure(1)
        i = 0
        for coher in self.cohers:
            i += 1
            plt.subplot(nRows, nCols, i)
            plt.semilogx(self.freqs, np.absolute(coher.data))
            plt.title(_seed_code(self.stats[coher.chan_nums[0]]) + '-' +
                      _seed_code(self.stats[coher.chan_nums[1]]))
            plt.ylim([0, 1])
        if outfile:
            plt.savefig(outfile)
        else:
            plt.show()
        return


class TransferFunction:
    def __init__(self, freqs, data, uncerts, drive_stats, resp_stats,
                 gooddata=None, noisechan='response'):
        """
        :parm freqs: 1-D array of frequencies
        :parm data: 1-D array of transfer function
        :parm uncerts: 1-D array of uncertainties
        :parm units: data units (str)
        :parm noisechan: 'response', 'driving', 'equal' or 'unknown'
        :type drive_stats, resp_stats: :class:`~obspy.core.trace.Stats`
        """
        self.freqs = freqs
        self.data = data
        self.uncerts = uncerts
        self.drive_stats = drive_stats
        self.resp_stats = resp_stats
        self.gooddata = gooddata
        self.noisechan = noisechan

    @classmethod
    def calc(cls, spects, cohers, drivechan='H', respchan='Z',
             noisechan='response', verbose=True):
        """
        Calculate the transfer function between 2 channels

        XF = calcXF(spect,cohers,drivechan,respchan,noisechan)
        input:
            spects    (list)  contains PSDs calculated using calc_PSDs()
            cohers    (list)  contains coherences
            drivechan (str)   the driving channel name (or last character(s))
            respchan  (str)   the response channel name (or last character(s))
            noisechan (str)	  is which channel to assume contains the noise
                'response'	[default] Assume all noise on the response channel
                'driving'	Assume all noise on the driving channel
                'equal'		Assume the same signal/noise on the driving  and
                            response channels
                'unknown'	Make no assumption about noise
        output:
            dictionary containing the transfer function
        """

        if verbose:
            print(f'Calculating XF between "{drivechan}" (driving) and'
                  f'"{respchan}" (response) channels, assume noise is on'
                  f'{noisechan}')

        # FIND PSD and coherence channels matching drivechan and respchan
        coh = None
        for c in cohers:
            chan_i = cohers.stats[c.ch_nums[0]].channel
            chan_j = cohers.stats[c.ch_nums[1]].channel
            if (chan_i.endswith(drivechan) and chan_j.endswith(respchan)) or \
               (chan_i.endswith(respchan) and chan_j.endswith(drivechan)):
                coh = c
                break
        if not coh:
            warnings.warn('Did no t find a coherence with channels {}, {}'.
                          format(drivechan, respchan))
            return False
        drive_spect = None
        resp_spect = None
        for s in spects:
            if s['chan'].endswith(drivechan):
                drive_spect = s
                if verbose:
                    print('drive_spect channel is "{}"'.format(s['chan']))
            elif s['chan'].endswith(respchan):
                resp_spect = s
                if verbose:
                    print('resp_spect channel is "{}"'.format(s['chan']))
        if not drive_spect:
            warnings.warn('Did not find a spectra with channel {}'.
                          format(drivechan))
            return False
        if not resp_spect:
            warnings.warn('Did not find a spectra with channel {}'.
                          format(respchan))
            return False

        # CALCULATE RESP/DRIVE
        RespOverDrive = np.sqrt(np.divide(resp_spect['data'],
                                          drive_spect['data']))

        # CALCULATE TRANSFER FUNCTION
        # Equations from Bendat&Piersol "Random Data" 1986, pp 176-181 (xfs)
        # & pp 317, Table 9.6
        cohmagsq = np.multiply(np.absolute(coh['data']),
                               np.absolute(coh['data']))
        errbase = np.divide(np.sqrt(np.ones(cohmagsq.shape) - cohmagsq),
                            2 * coh['num_windows'] * cohmagsq)
        if noisechan == 'response':
            xf = np.multiply(RespOverDrive, coh['data'])
            xferr = np.multiply(np.abs(xf), errbase)
        elif noisechan == 'driving':
            xf = np.divide(RespOverDrive, coh['data'])
            xferr = np.multiply(np.abs(xf), errbase)
        elif noisechan == 'equal':
            xf = RespOverDrive
            xferr = np.abs(np.multiply(xf, errbase))
        elif noisechan == 'unknown':
            xf = RespOverDrive
            # Ad-hoc error guesstimate
            maxerr = np.abs(np.power(coh['data'], -1.)) + errbase
            minerr = np.abs(coh['data']) - errbase
            xferr = np.abs(np.multiply(xf, (maxerr-minerr) / 2))
        else:
            warnings.warn('Invalid: noisechan = ' + noisechan)
            return False
        return cls(drive_spect.freqs, xf, xferr,
                   drive_spect.stats, resp_spect.stats,
                   np.absolute(coh['data']) > coh['signif_level'],
                   noisechan)
        # XF=dict(drive_info=dict(channel=drive_spect['chan'],
        #                         units=drive_spect['units']),
        #         resp_info= dict(channel=resp_spect['chan'],
        #                         units=resp_spect['units']),
        #         freq=     drive_spect['freq'],
        #         data=      xf,
        #         error=     xferr,
        #         noisechan=noisechan,
        #         gooddata=  np.absolute(coh['data']) > coh['signif_level'])

    def plot(self, outfile=None, debug=False):
        """
        plot transfer function
        """
        plt.figure(1)
        plt.clf()
        if debug:
            print(self.freqs[0])
            print(self.data[0])
            print(self.uncert[0])
        # plt.errorbar(np.log10(XF['freq']), np.absolute(XF['data']),
        #              yerr=np.absolute(XF['error']))
        # plt.ylim(0 , np.max(np.absolute(XF['data'])))
        plt.loglog(self.freqs, np.absolute(self.data), marker='o',
                   linestyle='')
        plt.loglog(self.freqs, np.absolute(self.data) + self.uncert)
        plt.loglog(self.freqs, np.absolute(self.data) - self.uncert)
        plt.ylim(np.min(np.absolute(self.data)),
                 np.max(np.absolute(self.data)))
        plt.title(f'Transfer function, noise channel: ({self.noisechan})')
        plt.ylabel(self.resp_stats.channel + '/' + self.drive_stats.channel)
        plt.xlabel('log(freq)')
        if outfile:
            plt.savefig(outfile)
        else:
            plt.show()
        return


def _fft_taper(data):
    """
    Cosine taper, 10 percent at each end (like done by [McNamara2004]_).

    .. warning::
        Inplace operation, so data should be float.
    """
    data *= cosine_taper(len(data), 0.2)
    return data


def _seed_code(stats):
    """
    Returns SEED code from obspy stats container

    >>> from obspy.core.trace import Stats
    >>> stats = Stats()
    >>> stats.network = 'YV'
    >>> stats.station = 'TEST'
    >>> stats.channel = 'BHZ'
    >>> _seed_code(stats)
    'YV.TEST..BHZ'
    """
    try:
        return '{}.{}.{}.{}'.format(stats.network, stats.station,
                                    stats.location, stats.channel)
    except AttributeError:
        return ''


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Plot LCHEAPO test results

Reads parameters from a lctest.yaml file
Plots:
    - time_series of one or more stations/channels
    - stacks of time series from one station/channel
    - spectra from  multiple stations/channels
    - particle_motions between two channels

There are 4 sections in each lctest.yaml file:

- ``input``: input data parameters
    - ``datafiles``: a list of the LCHEAPO files to read, each item contains:
        - ``name``: the filename
        - ``obs_type``: the obs type (for spectra plots).  possible values are
                        given in the help for ``lcplot``
        - ``station``: station name to use for all channels of this file

- ``output``: output plot file parameters
    - ``show``: show the plots?  If False, just save them to files
    - ``filebase``: start of each output filename (may include directory)

- ``plots``: the plots to make
    - ``time_series``: a list of waveform plots to make, each item contains:
        - ``description``: plot title
        - ``select``: parameters to select a subset of the
                      waveforms (see obspy.core.stream.Stream.select())
        - ``start_time``: plot start time
        - ``end_time``: plot end time

    - ``spectra``: list of spectra plots to make, each item contains:
        - ``description``: plot title
        - ``select``: as in ``time_series``
        - ``start_time``: data start time
        - ``end_time``: data end time
        - ``overlay``: overlay spectra on one plot? (True)

    - ``stack``: a list of stacked waveform plots to make.
               Useful when you performed a test action (tap, lift, jump, etc)
               several times.
               Each item contains:
        - ``description``: plot title
        - ``components``: a list of orientation codes to plot (one plot
                                 for each orientation code)
        - ``times``: list of times to plot at (each one "yyyy-mm-ddTHH:MM:SS")
        - ``offset_before.s``: start plot this many seconds before each `time`
        - ``offset_after.s``: end plot this many seconds after each ``time``

    - ``particle_motion``: a list of particle motion plots to make.
                           Use to evaluate the orientation and polarity of
                           the channels.
                           Each item contains:
        - ``description``**: as in ``time_series``
        - ``component_x``: component to plot on the x axis
        - ``component_y``: component code to plot on the y axis
        - ``times``: as in ``stack```
        - ``particle_offset_before.s``: start particle motion plot this many
                                        seconds before each ``time``
        - ``particle_offset_after.s``: end particle motion plot this many
                                       seconds after each ``time``
        - ``offset_before.s``: start time series plot this many seconds before
                               each ``time``
        - ``offset_after.s``: end time series plot this many seconds after
                              each ``time``

``plot_globals`` [optional]: Default values for each type of plot.
                             Same parameters as for ``plots``

Should allow ``lctest --examples`` option, to put example files in the
current directory
"""
import numpy as np
import yaml
import sys
import re
import argparse
import pkg_resources
import os
import json
from pathlib import Path

import jsonref
import jsonschema
from matplotlib import pyplot as plt
from obspy.core import UTCDateTime, Stream
from obspy.signal import PPSD
# from crawtools.spectral import SpectralDensity  # PSDs
from tiskitpy import SpectralDensity  # PSDs

from .lcread import read as lcread


def main():
    """
    Read the yaml file and plot the specified tests
    """
    args = get_arguments()
    root = read_lctest_yaml(args.yaml_file)
    plot_globals = get_plot_globals(root)
    show = root['output']['show']
    filebase = root['output']['filebase']
    inputs = root['input']
    if 'time_series' in root['plots']:
        for plot_info in root['plots']['time_series']:
            plot_info = _add_defaults(
                plot_info, plot_globals.get('time_series', None))
            plot_time_series(inputs, plot_info, filebase, show, args.verbose)
    if 'spectra' in root['plots']:
        for plot_info in root['plots']['spectra']:
            plot_info = _add_defaults(
                plot_info, plot_globals.get('spectra', None))
            plot_spect(inputs, plot_info, filebase, show)
    if 'stack' in root['plots']:
        for plot_info in root['plots']['stack']:
            plot_info = _add_defaults(
                plot_info, plot_globals.get('stack', None))
            start_time_str, end_time_str = _get_time_limits(plot_info)
            stream = _read_inputs(inputs, start_time_str, end_time_str)
            if 'select' in plot_info:
                stream = stream.select(**plot_info['select'])
            if stream is None:
                break
            for o_code in plot_info['components']:
                for trace in _stream_component(stream, o_code):
                    plot_stack(trace, plot_info, filebase, show)
    if 'particle_motion' in root['plots']:
        for plot_info in root['plots']['particle_motion']:
            plot_info = _add_defaults(
                plot_info, plot_globals.get('particle_motion', None))
            start_time_str, end_time_str = _get_time_limits(plot_info)
            stream = _read_inputs(root['input'], start_time_str, end_time_str)
            if 'select' in plot_info:
                stream = stream.select(**plot_info['select'])
            if stream is None:
                break
            streamx = _stream_component(stream, plot_info['component_x'])
            streamy = _stream_component(stream, plot_info['component_y'])
            for station in [t.stats.station for t in streamx]:
                plot_particle_motion(streamx.select(station=station)[0],
                                     streamy.select(station=station)[0],
                                     plot_info, filebase, show)


def plot_spect(input_list, plot_info, filebase, show):
    """
    Calculate and plot spectra for the input data

    Args:
        input_list (list of dict): input file information
        plot_info (dict): `spect` subelements
    """
    # Select appropriate channels
    title = plot_info["description"]
    print(f'Plotting spectra, {title}')
    stream = _read_inputs(input_list, plot_info.get('start_time'),
                          plot_info.get('end_time'))
    if stream is None:
        return
    if 'select' in plot_info:
        stream = stream.select(**plot_info['select'])
    # Calculate spectra
    wl = plot_info.get('window_length.s', None)
    # wl = _get_val('window_length.s', plot_info, glob_parms)
    if wl:
        spect = SpectralDensity.from_stream(stream, window_s=wl)
    else:
        spect = SpectralDensity.from_stream(stream)

    # Plot spectra
    overlay = plot_info.get('overlay', True)
    if filebase:
        outfile = _make_plot_filename(filebase, plot_info.get('select', None),
                                      'spectra', title)
    else:
        outfile = None
    spect.plot(outfile=outfile, overlay=overlay, show=show, title=title)


def plot_time_series(input_list, plot_info, filebase, show, verbose=False):
    """
    Plot a time series

    Args:
        input_list (list of dict): input file information
        plot_info (dict): information about the current plot
        filebase (str): file basename
        show (bool): show plot on screen
        verbose (bool): be verbose
    """
    print('Plotting time series "{}"'.format(plot_info["description"]))
    title = plot_info["description"]
    stream = _read_inputs(input_list, plot_info.get('start_time'),
                          plot_info.get('end_time'))
    if stream is None:
        return
    if verbose:
        print(stream)
    if 'select' in plot_info:
        stream = stream.select(**plot_info['select'])
    outfile = None
    if filebase:
        outfile = filebase

    fig = plt.figure()
    fig.suptitle(title)

    if filebase:
        outfile = _make_plot_filename(filebase, plot_info.get('select', None),
                                      'ts', title)
        stream.plot(size=(800, 600), fig=fig, outfile=outfile)
    if show:
        stream.plot(size=(800, 600), fig=fig)
        plt.show()


def plot_stack(trace, plot_info, filebase, show):
    """
    Plot a stack of time series from one trace

    Args:
        trace is an obspy Trace object
        plot_info (dict): information about the current plot
        filebase (str): base filename to write the plot to (None)
        show (bool): show plot on screen
    """
    title = plot_info['description'] + f', {trace.get_id()}'
    print(f'Plotting stack "{title}"')
    times = [UTCDateTime(t) for t in plot_info['times']]
    offset_before = plot_info.get('offset_before.s', None)
    offset_after = plot_info.get('offset_after.s', None)
    plot_span = plot_info.get('plot_span', None)
    print(f'{offset_before=}, {offset_after=}, {plot_span=}')
    assert offset_before >= 0,\
        f'plot_stack "{title}": offset_before < 0 ({offset_before:g})'
    assert offset_after > 0,\
        f'plot_stack "{title}": offset_after <= 0 ({offset_after:g})'
    if plot_span:
        _plot_span(times, Stream(traces=[trace]))
    colors = plt.cm.rainbow(np.linspace(0, 1, len(times)))
    offset_vertical = 0
    # time_zero = UTCDateTime(times[0])
    ax = plt.axes()
    max_val = 0
    # Set up y axis range
    for time in times:
        temp = trace.slice(time - offset_before, time + offset_after)
        if abs(temp.max()) > max_val:
            max_val = abs(temp.max())
    # Plot the subtraces
    for time, c in zip(times, colors):
        offset_vertical += max_val
        t = trace.slice(time - offset_before, time + offset_after)
        ax.plot(t.times("utcdatetime") - time,
                t.data - offset_vertical, color=c,
                label=time.strftime('%H:%M:%S') +
                '.{:02d}'.format(int(time.microsecond / 1e4)))
        offset_vertical += max_val
    ax.set_title(title)
    ax.grid()
    ax.legend()
    if filebase:
        outfile = _make_plot_filename(filebase, plot_info.get('select', None),
                                      'stack', title)
        plt.savefig(outfile)
    if show:
        plt.show()


def plot_particle_motion(tracex, tracey, plot_info, filebase, show):
    """
    Plot particle motions

    Args:
        tracex (:class:`obspy.core.stream.Stream): to plot on x axis
        tracey (:class:`obspy.core.stream.Stream): to plot on y axis
        plot_info (dict): information about the current plot
        filebase (str): base filename to write the plot to (None)
        show (bool): show plot on screen
    """
    description = plot_info['description']
    title = '{}, {} vs {}'.format(description, tracex.get_id(),
                                  tracey.get_id())
    print(f'Plotting particle motion "{title}"')
    times = [UTCDateTime(t) for t in plot_info['times']]
    offset_before = plot_info.get('particle_offset_before.s', None)
    offset_after = plot_info.get('particle_offset_after.s', None)
    offset_before_ts = plot_info.get('offset_before.s', None)
    offset_after_ts = plot_info.get('offset_after.s', None)
    plot_span = plot_info.get('plot_span', None)
    if plot_span:
        _plot_span(times, Stream(traces=[tracex, tracey]))
    # Setup axis grid
    fig = plt.figure()
    gs = fig.add_gridspec(2, 3, hspace=0, wspace=0)
    # two columns, one row:
    axx = fig.add_subplot(gs[0, :2])
    axy = fig.add_subplot(gs[1, :2], sharex=axx, sharey=axx)
    # one column, one row
    axxy = fig.add_subplot(gs[1, 2], sharey=axy)
    tx_comp = tracex.stats.channel[-1]
    ty_comp = tracey.stats.channel[-1]
    for time in times:
        # time series plots
        _plot_one_ts(axx, tracex, tx_comp, time,
                     offset_before, offset_after,
                     offset_before_ts, offset_after_ts)
        _plot_one_ts(axy, tracey, ty_comp, time,
                     offset_before, offset_after,
                     offset_before_ts, offset_after_ts)

        # partical motion plot
        tx = tracex.slice(time - offset_before, time + offset_after)
        ty = tracey.slice(time - offset_before, time + offset_after)
        axxy.plot(tx.data, ty.data)
        axxy.axvline(0)
        axxy.axhline(0)
        axxy.set_aspect('equal', 'datalim')
        axxy.set_xlabel(tx_comp)
        # axxy.set_yticklabels([])
    fig.suptitle(title)
    if filebase:
        outfile = _make_plot_filename(filebase, plot_info.get('select', None),
                                      'pm', title)
        plt.savefig(outfile)
    if show:
        plt.show()


def _stream_component(stream, component):
    """ Return the traces corresponding to the given component """
    try:
        return stream.select(component=component)
    except IndexError:
        print(f'Did not find component {component}')
        sys.exit()


def _get_time_limits(plot_info):
    """
    Get time limits for an object containing offsets and a list of `times`

    Args:
        plot_info (dict): dictionary containing ``times``, `offset_before.s`
                          and `offset_after.s` elements
    """
    max_offset_before = plot_info['offset_before.s']
    if 'offset_before_ts.s' in plot_info:
        max_offset_before = max(max_offset_before,
                                plot_info['offset_before_ts.s'])
    max_offset_after = plot_info['offset_after.s']
    if 'offset_after_ts.s' in plot_info:
        max_offset_after = max(max_offset_after,
                               plot_info['offset_after_ts.s'])

    min_time = UTCDateTime(plot_info['times'][0]) - max_offset_before
    max_time = UTCDateTime(plot_info['times'][0]) + max_offset_after
    if len(plot_info['times']) > 1:
        for time_str in plot_info['times'][1:]:
            time_obj = UTCDateTime(time_str)
            if time_obj - max_offset_before < min_time:
                min_time = time_obj - max_offset_before
            if time_obj + max_offset_after > max_time:
                max_time = time_obj + max_offset_after
    return min_time.isoformat(), max_time.isoformat()


def _read_inputs(input_list, start_time_str, end_time_str):
    """
    Read selected time range from input files

    Args:
        input_list: list of files to read and their characteristics
        start_time_str (str): string representation of start time
        end_time_str (str): string representation of end time
    """
    try:
        starttime = UTCDateTime(start_time_str)
        endtime = UTCDateTime(end_time_str)
    except Exception:
        print('start_ or end_time badly formatted')
        return None
    return read_files(input_list, starttime, endtime)


def _make_plot_filename(filebase, selectargs, plot_type, type_info,
                        suffix='.png'):
    """
    Makes a valid filename from the list of elements

    Args:
        filebase (str): start of filename (may include directories)
        selectargs (dict): arguments provided to "select()""
        plot_type (str)): 'spectra', 'ts', 'stack', 'pm'
        type_info (str)): information specific to the plot type
    """
    filepath = Path(filebase)
    filebase = filepath.name
    fileparent = filepath.parent
    fileparent.mkdir(exist_ok=True)
    if selectargs is not None:
        select_strs = '_'.join([key+arg for key, arg in selectargs.items()])
    else:
        select_strs = ''
    cleaned = [_get_valid_filename(e) for e in (filebase, plot_type,
                                                select_strs, type_info)]
    assert suffix[0] == '.'
    return str(fileparent / ('_'.join(cleaned) + suffix))


def _plot_one_ts(ax, trace, comp, time, offset_before_pm, offset_after_pm,
                 offset_before, offset_after):
    t = trace.slice(time - offset_before, time + offset_after)
    ax.plot(t.times("utcdatetime") - time, t.data)
    ax.axvline(-offset_before_pm, color='k', linestyle='--')
    ax.axvline(offset_after_pm, color='k', linestyle='--')
    ax.set_ylabel(comp)


def plot_PPSD(trace, sta, start_time, interval=7200, filebase=None,
              show=True):
    """
    Plot a Probabilistic Power Spectral Desnsity for the trace

    trace = obspy Trace objet
    sta = obspy Inventory/Station object corresponding to the trace
    start_time = time at which to start spectra
    interval=offset between PSDs (seconds, minimum=3600)
    """
    now_time = trace.start_time
    first_read = True
    while now_time < trace.end_time-interval:
        if first_read:
            if trace.stats.component[1] == 'D':
                ppsd = PPSD(trace.stats, metadata=sta,
                            special_handling='hydrophone')
            else:
                ppsd = PPSD(trace.stats, metadata=sta)
            first_read = False
        ppsd.add(trace)
        now_time += interval

    ppsd.save_npz(f'{filebase}_PPSD.npz')
    if filebase:
        description = '{}.{}.{}.{}'.format(trace.stats.network,
                                           trace.stats.station,
                                           trace.stats.location,
                                           trace.stats.channel)
        ppsd.plot(filebase + '_' + description
                  + '_PPSD.png')
    if show:
        plt.plot()
    # ppsd.plot_temporal([0.1,1,10])
    # ppsd.plot_spectrogram()
    return 0


def _UTCDateTimeorNone(input):
    """
    Converts input string to UTCDateTime or None if it doesn't work
    """
    try:
        return UTCDateTime(input)
    except TypeError:
        return None


def _add_defaults(localdict, defaultdict):
    """
    Returns localdict, completed by defaultdict

    If a key is in defaultdict and not in localdict, add to localdict
    """
    if defaultdict is None:
        return localdict
    for k, v in defaultdict.items():
        if k not in localdict:
            localdict[k] = v
    return localdict


def get_plot_globals(root):
    """
    Return global plot values
    """
    globals = root.get('plot_globals', None)
    return globals


def read_files(inputs, starttime=None, endtime=None, verbose=False):
    """
    Read in data from one or more datafiles

    datafile_list: list of dict(name=, obs_type=, station=)
    starttime (:class `obspy.UTCDateTime`): data start time
    endtime (:class `obspy.UTCDateTime`): data end time
    """
    if verbose:
        if len(inputs['datafiles']) == 1:
            print(f"Reading file {inputs['datafiles'][0]['name']}")
        else:
            print(f"Reading files {[x['name'] for x in inputs['datafiles']]}")
    stream = Stream()
    for df in inputs['datafiles']:
        s = lcread(df['name'], starttime=starttime, endtime=endtime,
                   obs_type=df['obs_type'])
        if s is not None:
            for t in s:
                t.stats.station = df['station']
            stream += s
        elif verbose:
            print('No data between {} and {} in file {}'.format(
                starttime.isoformat(), endtime.isoformat(), df['name']))
    if len(stream) > 0:
        return stream
    else:
        print('No data between {} and {} found in any file'.format(
            starttime.isoformat(), endtime.isoformat()))
        return None


def get_arguments():
    """
    Get command line arguments
    """
    p = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    p.add_argument('yaml_file', help='YAML parameter file')
    p.add_argument("-v", "--verbose", default=False, action='store_true',
                   help="verbose")
    p.add_argument("--inv_file", help="inventory file")
    return p.parse_args()


def _get_valid_filename(s):
    assert isinstance(s, str)
    s = s.strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)


def _plot_span(times, stream):
    first = min(times)
    last = max(times)
    span = last-first
    stream.slice(first - span / 2, last + span / 2).plot()


def read_lctest_yaml(filename):
    """
    Verify and read in an lctest yaml file
    """
    with open(filename) as f:
        root = yaml.safe_load(f)
        if not validate_schema04(root, 'lctest'):
            sys.exit()
    return root


def validate_schema04(instance: dict, type: str = 'lctest'):
    """
    Validates a data structure against a draft 04 jsonschema

    :param instance: data structure read from a yaml or json filebase
    :param type: type of the data structure
    """
    schema_file = pkg_resources.resource_filename(
        "lcheapo", f"data/{type}.schema.json")
    base_path = os.path.dirname(schema_file)
    base_uri = f"file:{base_path}/"
    with open(schema_file, "r") as f:
        try:
            schema = jsonref.loads(f.read(), base_uri=base_uri,
                                   jsonschema=True)
        except json.decoder.JSONDecodeError as e:
            print(f"JSONDecodeError: Error loading schema file: {schema_file}")
            print(str(e))
            return False
        # except:
        #     print(f"Error loading JSON schema file: {schema_file}")
        #     print(sys.exc_info()[1])
        #     return False

    # Lazily report all errors in the instance
    # ASSUMES SCHEMA IS DRAFT-04
    try:
        v = jsonschema.Draft4Validator(schema)

        if not v.is_valid(instance):
            print("")
            for error in sorted(v.iter_errors(instance), key=str):
                print("\t\t", end="")
                for elem in error.path:
                    print(f"['{elem}']", end="")
                print(f": {error.message}")
            print("\tFAILED")
            return False
        else:
            print("Schema validated")
            return True
    except jsonschema.ValidationError as e:
        print("")
        print("\t" + e.message)
        return False


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Create/append process-steps.json file
"""
from pathlib import Path
from dataclasses import dataclass, field
import json
import argparse
import copy
from datetime import datetime as dt


@dataclass
class ProcessStep:
    """
    Create an SDPChain process step
    
    Arguments:
        app_name (str): the application name
        cmdline (str): the full command line
        
    Attributes:
        date(datetime.datetime): datetime at which the command was called
        exit_status (int): command exit status
        app_description (str)
        app_version (str)
        parameters (dict or argparse.NameSpace): execution parameters
        messages (list of str): execution messages
        tools (list of str): tools used by the application
        output_files (list of str): files output by the program (and not
            in parameters)
        output_file (str): file output during execution (and not in parameters)
    """
    app_name: str
    cmdline: str
    date: dt = dt.utcnow()
    exit_status: int = None
    app_description: str = 'No description found'
    app_version: str = "Unknown"
    parameters: dict = field(default_factory=dict)
    messages: list = field(default_factory=list)
    tools: list = field(default_factory=list)
    output_files: list = field(default_factory=list)
    output_file: str = ''

    def __post_init__(self):
        # If parameters provided by argparser
        if isinstance(self.parameters, argparse.Namespace):
            self.parameters = copy.deepcopy(vars(self.parameters))

    def log(self, text, write_to_screen=False):
        """
        add text to messages list
        """
        if write_to_screen:
            print(text)
        self.messages.append(text)

    def write(self, in_dir, out_dir, filename='process-steps.json',
              quiet=False, verbose=False):
        """
        Write the Process Step to a process-steps file

        Args:
            in_dir: directory containing input process-steps file
            out_dir: directory in which to create out process-steps file
            quiet (bool): don't write anything out
            verbose (bool): print out lots of stuff
        """
        if verbose==True:
            quiet=False
        self._modify_parameters()
        step = {'application': dict(name=self.app_name,
                                    description=self.app_description,
                                    version=self.app_version),
                'execution': dict(command_line=self.cmdline,
                                  date=self.date.strftime('%Y-%m-%dT%H:%M:%S'),
                                  messages=self.messages,
                                  parameters=self.parameters,
                                  tools=self.tools,
                                  exit_status=self.exit_status)
        }
 
        # READ FILE FROM INPUT DIRECTORY
        in_file = Path(in_dir) / filename
        out_file = Path(out_dir) / filename
        if verbose is True:
            print(f'Writing to {str(out_file)}')
        tree = {}
        try:
            fp = open(in_file, "r")
        except FileNotFoundError:  # File not found
            pass
        else:   # File found
            if verbose is True:
                print(f'\tFile exists, will append')
            try:
                tree = json.load(fp)
            except Exception:
                # Couldn't read input file, don't overwrite it either
                if in_file == out_file:
                    out_file = _unique_path(Path(out_dir),
                                                'process-steps{:02d}.json')
                if not quiet:
                    print('{} is unreadable. Writing to new file {}, which will lack previous steps'
                          .format(in_file, out_file))
            fp.close()
        if 'steps' in tree:
            tree['steps'].append(step)
        else:
            tree['steps'] = [step]
        # WRITE FILE TO OUTPUT DIRECTORY
        if out_file.exists():
            out_file = _unique_path(Path(out_dir),
                                    'process-steps{:02d}.json')
        with open(out_file, "w") as fp:
            json.dump(tree, fp, sort_keys=True, indent=4)

    def _modify_parameters(self):
        """
        Modify execution parameters to conform to schema
        
        puts parameters['in_dir', 'out_dir', base_dir] in 
        parameters['directory_paths']{['input'], ['output'], 'base'}
        
        puts output_files in parameters['output_files']
        puts output_file in parameters['output_file']
        """
        
        ep = self.parameters   # make a shortcut
        # base_dir, in_dir and out_dir all go into directory_paths
        if 'base_dir' in ep or 'in_dir' in ep or 'out_dir' in ep:
            ep['directory_paths'] = {}
            dp = ep['directory_paths']
            if 'base_dir' in ep:
                dp['base'] = ep['base_dir']
                del ep['base_dir']
            if 'in_dir' in ep:
                dp['input'] = ep['in_dir']
                del ep['in_dir']
            if 'out_dir' in ep:
                dp['output'] = ep['out_dir']
                del ep['out_dir']
        
        # output_files go into parameters['output_files']
        if self.output_files:
            if 'output_files' in ep:
                ep['output_files'] = ep['output_files'].extend(self.output_files)
            else:
                ep['output_files'] = self.output_files
        if self.output_file:
            if 'output_file' in ep:
                ep['output_files'] = [ep['output_file'], self.output_file]
                del ep['output_file']
            else:
                ep['output_file'] = self.output_file
    
    @staticmethod
    def setup_paths(args, expand_wildcards=True, verbose=True):
        """
        Set up paths using SDPCHAIN standards
    
        Args:
            args (:class:NameSpace): usually created by argparser, has
                    attributes base_dir, in_dir, out_dir and input_files
            expand_wildcards (bool): expand captured wildcards in input_files?
        Returns:
            (tuple):
                in_dir, out_dir, input_files: base_dir-adjusted paths and filenames
        
        The rules are:
            - base_dir is the root for in_dir and out_dir)
            - in_dir is the directory for input files.  An absolute path or relative
                       to base_dir
            - out_dir is the directory to output to.
            - in_dir and out_dir are absolute paths or relative to base_dir
        """
        if not hasattr(args, "base_dir"):
            raise NameError('args has no base_dir attribute')

        # Work on in_dir
        if not hasattr(args, "in_dir"):
            raise NameError('args has no in_dir attribute')
        in_path = _choose_path(args.base_dir, args.in_dir)
        assert Path(in_path).is_dir() is True

        # Work on input_files
        if not hasattr(args, "input_files"):
            raise NameError('args has no input_files attribute')
        if expand_wildcards is True:
            input_files = [x.name for f in args.input_files
                           for x in Path(in_path).glob(f)]
        else:
            input_files = args.input_files
            
        # Work on out_dir
        if not hasattr(args, "out_dir"):
            return in_path, None, input_files
        out_path = _choose_path(args.base_dir, args.out_dir)
        assert not Path(out_path).is_file()
        if Path(out_path).exists() is False:
            if verbose:
                print(f"out_dir '{out_path}' does not exist, creating...")
            Path(out_path).mkdir(parents=True)

        return in_path, out_path, input_files
    
def _choose_path(base_dir, sub_dir):
    """ Set up absolute path to sub-directory """
    if Path(sub_dir).is_absolute():
        return sub_dir
    return str(Path(base_dir) / sub_dir)
    
    
def _unique_path(directory, name_pattern):
    counter = 0
    while True:
        counter += 1
        path = directory / name_pattern.format(counter)
        if not path.exists():
            return path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
print out blocks with bad time header values
"""
from lcheapo import (LCDataBlock, LCDiskHeader)

fname = '../tests/data/BUGGY.raw.lch'


def _read_and_print_dataheader(fp, block_num, title):
    lcData = LCDataBlock()
    lcData.seekBlock(fp, block_num)
    lcData.readBlock(fp)
    print(f'{title:12s} = {block_num:10d}: ', end='')
    lcData.prettyPrintHeader(annotated=True)


with open(fname, 'rb') as fp:
    # Get first data block number from the header
    lcHeader = LCDiskHeader()
    lcHeader.seekHeaderPosition(fp)
    lcHeader.readHeader(fp)
    first_data_block = lcHeader.dataStart
    # first_data_block = 0   # Force to start in header (bad "headers")
    # lcHeader.printHeader()

    # Figure out last data block number
    fp.seek(0, 2)                # Seek end of file
    last_data_block = int(fp.tell() / 512) - 1

    # Print out first and last data blocks
    _read_and_print_dataheader(fp, first_data_block, 'First block')
    _read_and_print_dataheader(fp, last_data_block, 'Last block')

    # Run through data blocks, printing out ones with bad times
    lcData = LCDataBlock()
    lcData.seekBlock(fp, first_data_block)
    n_bad = 0
    for i in range(first_data_block, last_data_block + 1):
        lcData.readBlock(fp)
        if lcData.year > 50:     # > 50 is 1950-1999
            n_bad += 1
            print(f'{i:10d}: ', end='')
            lcData.prettyPrintHeader(annotated=True)
    print(f'{n_bad:d} bad header times')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = {}
with open("lcheapo/version.py") as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="lcheapo",
    version=version['__version__'],
    author="Wayne Crawford",
    author_email="crawford@ipgp.fr",
    description="LCHEAPO data routines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/WayneCrawford/lcheapo",
    packages=setuptools.find_packages(),
    include_package_data=True,
    # packages=['lcheapo'],
    # package_dir={'lcheapo': 'lcheapo'},
    # package_data={'lcheapo': ['data/*.xml', 'data/*.json',
    #                                 '_examples/*.py', '_examples/*.yaml']},
    install_requires=[
          'obspy>=1.1',
          'pyyaml>5.0',
          'jsonschema>=2.6',
          'jsonref>=0.2',
          'progress>=1.5',
          'tiskitpy>=0.4'
      ],
    entry_points={
         'console_scripts': [
             'sdpcat=lcheapo.sdpchain:sdpcat',
             'sdpstep=lcheapo.sdpchain:sdpstep',
             'lcfix=lcheapo.lcfix:main',
             'lcdump=lcheapo.lcdump:main',
             'lccut=lcheapo.lccut:main',
             'lcinfo=lcheapo.lcinfo:main',
             'lcheader=lcheapo.lcheader:main',
             'lcplot=lcheapo.lcplot:main',
             'lc2SDS_weak=lcheapo.lc2SDS:lc2SDS',
             'lc2ms_weak=lcheapo.lc2ms:lc2ms',
             'lctest=lcheapo.lctest:main',
             'lc_examples=lcheapo.lcputexamples:main'
         ]
    },
    python_requires='>=3.8',
    classifiers=(
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics"
    ),
    keywords='oceanography, marine, OBS'
)



#!/usr/bin/env python3
"""
    Run a command-line program and save info to process-steps file
"""
import subprocess
import sys
# import os
import argparse
from datetime import datetime
import distutils.spawn

from .version import __version__
from .process_steps import ProcessStep

assert sys.version_info >= (3, 5)

help_epilog = """
Some recommended instances of sdpstep:
TO INSERT A POSITIVE LEAP-SECOND (61-second minutes) AT $LEAPTIME
    ($LEAPTIME format is 'YYYY,DDD,HH,MM,SS,FFFF' delimiters can be [,:.])
    sdp-process "msmod --timeshift -1 -ts $LEAPTIME -s -o OUTFILE INFILE"
    sdp-process "msmod --actflags 4,1 -tsc $LEAPTIME -tec $LEAPTIME -s -i OUTFILE"
TO INSERT A NEGATIVE LEAP-SECOND (no second 59) AT $LPTIME
    sdp-process "msmod --timeshift +1 -ts $LEAPTIME -s -o OUTFILE INFILE"
    sdp-process "msmod --actflags 5,1 -tsc $LEAPTIME -tec $LEAPTIME  -s -i OUTFILE"
TO MERGE SEVERAL MINISEED FILES INTO ONE:
    sdp-process "msmod -o toto.mseed INFILES"
TO CHANGE CHANNEL/LOCATION CODES SEVERAL MINISEED FILES INTO ONE:
    sdp-process "msmod -o toto.mseed INFILES"

For use in the Seismology Data Preparation (SDP) chain

If tool_line_cmd is empty quotes, will create an unfilled process step
"""
app_description = "Run a command-line tool and write process-step"
app_name = "sdpstep"


def main():
    """
    Run a command line and enter run information into process-steps file

    -i and -o only affect the process-steps file, as each tool can have
    its own way to specify input and output files

    Uses subprocess toolbox, so redirects (">") and pipes ("|") won't work
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=app_description,
        epilog=help_epilog)
    parser.add_argument('tool_cmd_line', action="store",
                        help='The command line to run')
    parser.add_argument('-c', action="store", dest="exec_comment",
                        help='Execution comment')
    parser.add_argument('--version', action="store_true", dest="print_version",
                        default=False,
                        help='Print the version number and quit')
    parser.add_argument("-d", "--directory", dest="base_dir",
                        default='.', help="Base directory for files")
    parser.add_argument("-i", "--input", dest="in_dir", default='.',
                        help="path to input files (abs, or rel to base)")
    parser.add_argument("-o", "--output", dest="out_dir", default='.',
                        help="path for output files (abs, or rel to base)")
    args = parser.parse_args()
    if args.print_version:
        print(f'version {__version__}')
        sys.exit(0)

    process_step = ProcessStep(app_name,
                               " ".join(sys.argv),
                               app_description = app_description,
                               app_version = __version__,
                               parameters=args)

    args.in_dir, args.out_dir = ProcessStep.setup_paths(args)

    # if args.base_dir:
    #     print('Changing working directory to {}'.format(args.base_dir))
    #     os.chdir(args.base_dir)

    if len(args.tool_cmd_line) == 0:
        process_step.exit_code=0,
        process_step.tools=["tool"]
    else:
        # RUN THE COMMAND LINE
        if ">" in args.tool_cmd_line or "|" in args.tool_cmd_line:
            raise ValueError('Cannot have ">" or "|" on tool_cmd_line')
        tool = args.tool_cmd_line.split()[0]
        if not is_tool(tool):
            raise OSError(f'tool "{tool}" not found')
        return_code, exec_message = run_command_line(args.tool_cmd_line)

        # WRITE INFORMATION TO PROCESS-STEPS FILE
        tool = args.tool_cmd_line.split()[0]
        params = dict(tool_version=_get_tool_version(tool),
                      tool_description=_get_tool_description(tool))
        process_step.exit_code=0,
        process_step.messages = exec_message
        process_step.tools=[tool]
    process_step.write(args.in_dir, args.out_dir)
    return


def is_tool(name):
    return distutils.spawn.find_executable(name) is not None


def run_command_line(command_line):
    """
    Runs the command line and returns information
    """
    try:
        # p = subprocess.check_output(command_line, stderr=subprocess.STDOUT)
        CP = subprocess.run(command_line.split(), check=True,
                            capture_output=True)   # ,
        # stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as err:
        return_code = err.returncode
        exec_message = str(err.output, encoding='utf-8',
                           errors='strict').splitlines()

    else:
        return_code = 0
        exec_message = str(CP.stdout, encoding='utf-8',
                           errors='strict').splitlines()
    return return_code, exec_message


def _get_tool_version(tool):
    """
    Get version information about the command

    :param tool: the command-line tool
    :type tool: str
    :returns: version information
    :rtype: str
    """
    option_list = ["--version", "-V", "", '-h']
    for option in option_list:
        outp = _get_test_output([tool, option])
        if outp:
            return _get_first_line(outp)
    return ''


def _get_tool_description(tool):
    """
    Get description of the command

    :param tool: the command-line tool
    :type tool: str
    :returns: version information
    :rtype: str
    """
    option_list = ["--help", "-h", ""]
    for option in option_list:
        outp = _get_test_output([tool, option])
        if outp:
            return _get_first_line(outp)
    return 'None'


def _get_test_output(cmd_list):
    """
    Get output of command for test.

    :param cmd_list:command and any arguments
    :type cmd_list: str
    returns: List of the text lines from stdout+stderr
             or None if none output or command fails
    """
    try:
        CP = subprocess.run(cmd_list, timeout=5,
                            check=True, capture_output=True,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except Exception:
        return None
    else:
        if CP.returncode == 0:
            outp = CP.stdout
            outp = str(outp, encoding='utf-8', errors='strict')
            if type(outp) == 'str':
                outp = outp.splitlines()
        else:
            return None
    return outp


def _get_first_line(inp):
    if isinstance(inp, str):
        return inp.split('\n')[0]
    if isinstance(inp, list):
        if isinstance(inp[0], str):
            return inp[0].split('\n')[0]
    return ""


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Data access for lcheapo files
"""
import sys
import datetime
import struct
# import string
import os

# ------------------------------------
# Global Variable Declarations
# ------------------------------------
PROGRAM_NAME = "lcheapo.py"
VERSION = "0.3.0"
HEADER_START = 2
BLOCK_SIZE = 512


class LCCommon:
    """
    LCHEAPO Parent Class
    """
    def __init__(self):
        (self.msec, self.second, self.minute, self.hour,
         self.day, self.month, self.year) = (0, 0, 0, 0, 1, 1, 2000)

    def fixYear(self, year):
        """
        Change year from 2 to 4 digits
        """
        if (0 <= year < 50):
            year += 2000
        elif (50 < year < 100):
            year += 1900
        return year

    def getDateTime(self):
        """
        Convert year:day:month:hour:minute:second.msec value to datetime
        """
        try:
            return datetime.datetime(self.fixYear(self.year),
                                     self.month, self.day,
                                     self.hour, self.minute,
                                     self.second, self.msec*1000)
        except ValueError:
            # return bogus date
            return datetime.datetime(1900, 1, 1, 0, 0, 0, 0)

    def changeTime(self, tm):
        """
        Change local values to correspond to datetime object
        """
        if (tm.year < 2000):
            self.year = tm.year - 1900
        else:
            self.year = tm.year - 2000

        self.month = tm.month
        self.day = tm.day
        self.hour = tm.hour
        self.minute = tm.minute
        self.second = tm.second
        self.msec = int(tm.microsecond / 1000)

    def getRealSampleRate(self, sampleRate):
        """
        Return floating point sample rate
        """
        if (sampleRate < 61):
            return 31.25
        elif (sampleRate < 125):
            return 62.50
        return float(sampleRate)

    def determineLastBlock(self, fp):
        """
        Determine the last full block in the file
        """
        currentPos = fp.tell()
        fp.seek(-1, 2)
        lastBlock = int(fp.tell() / BLOCK_SIZE)
        if currentPos % BLOCK_SIZE != 0:
            lastBlock -= 1
        fp.seek(currentPos, 0)
        return lastBlock

    def seekBlock(self, fp, block):
        """
        Go to a particular block in the file
        """
        fp.seek(block * BLOCK_SIZE, os.SEEK_SET)


class LCUserData (LCCommon):
    """
    LCheapo user data section class
    """
    def __init__(self):
        self.userData = ""

    def seekUserDataPosition(self, fp):
        """
        Move the file pointer to the user data position
        """
        fp.seek(0, os.SEEK_SET)

    def isValidData(self):
        if self.userData.find("LCHEAPO") != 0:
            return False
        else:
            return True

    def readData(self, fp):
        self.userData = fp.read(2048)

    def writeData(self, fp):
        # ------ NOT FINISHED ------
        fp.write(self.userData)


class LCDiskHeader (LCCommon):
    """
    LCheapo disk header class
    """
    def __init__(self):
        pass

    def fixDiskSizeBug(self):
        "Fix an old bug which doubled the disk size."
        self.diskSize /= 2

    def seekHeaderPosition(self, fp):
        "Move the file pointer to the standard header position."
        fp.seek(HEADER_START*BLOCK_SIZE, os.SEEK_SET)

    def readHeader(self, fp):
        """
        :returns: 1 if good read, 0 otherwise
        """
        self.seekHeaderPosition(fp)
        "Read an LCheapo disk header (packed big endian format)."
        (self.writeBlock, self.writeByte, self.readBlock, self.readByte) =\
            struct.unpack('>LHLH', fp.read(12))
        (self.dirStart, self.dirSize, self.dirBlock, self.dirCount) =\
            struct.unpack('>4L', fp.read(16))
        (self.slowStart, self.slowSize, self.slowBlock, self.slowByte) =\
            struct.unpack('>4L', fp.read(16))
        (self.logStart, self.logSize, self.logBlock, self.logByte) =\
            struct.unpack('>4L', fp.read(16))
        (self.dataStart, self.diskNumber) =\
            struct.unpack('>LH', fp.read(6))
        (self.softwareVersion, self.description) =\
            struct.unpack('>10s80s', fp.read(90))
        # Python strings do not terminate on '\0', therefore, do this manually.
        self.softwareVersion = _str_from_cstr(self.softwareVersion)
        if self.softwareVersion is None:
            print('Could not read softwareVersion, file may not have a header')
            return 0
        self.description = _str_from_cstr(self.description)

        (self.sampleRate, self.startChannel, self.numberOfChannels) =\
            struct.unpack('>3H', fp.read(6))
        (self.slowDataRate, self.slowStartChannel,
         self.slowNumberOfChannels) = struct.unpack('>3H', fp.read(6))
        (self.dataType, self.diskSize, self.ramSize, self.numberOfWindows) =\
            struct.unpack('>4H', fp.read(8))

        # Additions which cannot be written back out
        self.realSampleRate = self.getRealSampleRate(self.sampleRate)
        self.dataTypeString = "Unknown Compression"
        if self.dataType >= 0 and self.dataType <= 3:
            self.dataTypeString = ("Uncompressed (16-Bit)",
                                   "Compressed (16-Bit)",
                                   "Uncompressed (24-Bit)",
                                   "Compressed (24-Bit)")[self.dataType]
        return 1

    def writeHeader(self, fp):
        "Write an LCheapo disk header (packed big endian format)."
        fp.write(struct.pack('>LHLH', self.writeBlock, self.writeByte,
                             self.readBlock, self.readByte))
        fp.write(struct.pack('>4L', self.dirStart, self.dirSize, self.dirBlock,
                             self.dirCount))
        fp.write(struct.pack('>4L', self.slowStart, self.slowSize,
                             self.slowBlock, self.slowByte))
        fp.write(struct.pack('>4L', self.logStart, self.logSize,
                             self.logBlock, self.logByte))
        fp.write(struct.pack('>LH', self.dataStart, self.diskNumber))
        fp.write(struct.pack('>10s80s',
                             _cstr_from_str(self.softwareVersion),
                             _cstr_from_str(self.description)))
        fp.write(struct.pack('>3H', self.sampleRate, self.startChannel,
                             self.numberOfChannels))
        fp.write(struct.pack('>3H', self.slowDataRate, self.slowStartChannel,
                             self.slowNumberOfChannels))
        fp.write(struct.pack('>4H', self.dataType, self.diskSize,
                             self.ramSize, self.numberOfWindows))

    def printHeader(self):
        "Print the disk header information."
        print("\n\n")
        print("===========================")
        print("Disk Header")
        print("===========================\n")
        print("Field Name                 Contents")
        print("----------                 --------")
        f_int = "{:22s} {:12d} ({})"
        f_str = "{:35s} {}"
        print(f_int.format("Write Block", self.writeBlock,
                           "Next block to write"))
        print(f_int.format("Write Byte", self.writeByte, "Next byte to write"))
        print(f_int.format("Read Block", self.readBlock, "Unused"))
        print(f_int.format("Read Byte ", self.readByte, "Unused"))

        print("\n==== Dir Info ====")
        print(f_int.format("Dir Start", self.dirStart, "Dir start block"))
        print(f_int.format("Dir Size ", self.dirSize, "Max Dir entries"))
        print(f_int.format("Dir Block", self.dirBlock, "Next Dir block"))
        print(f_int.format("Dir Count", self.dirCount, "Total Dir entries"))

        print("\n==== Slow Data Storage ====")
        print(f_int.format("Slow Start", self.slowStart,
                           "Slow data start block)"))
        print(f_int.format("Slow Size ", self.slowSize,
                           "Slow data size in blocks"))
        print(f_int.format("Slow Block", self.slowBlock,
                           "Next block number to write"))
        print(f_int.format("Slow Byte ", self.slowByte,
                           "Next byte number to write"))

        print("\n==== Log Data Info ====")
        print(f_int.format("Log Start", self.logStart, "Log data start block"))
        print(f_int.format("Log Size ", self.logSize,  "Log size in blocks"))
        print(f_int.format("Log Block", self.logBlock,
                           "Next block number to write"))
        print(f_int.format("Log Byte ", self.logByte,
                           "Next byte number to write"))

        print("\n==== General Information ====")
        print(f_int.format("Data Start ", self.dataStart,
                           "Normal data start block"))
        print(f_int.format("Disk Number", self.diskNumber, ""))
        print(f_str.format("Software Version", self.softwareVersion))
        print(f_str.format("Description", self.description))
        print("{:22s} {:12.2f} Hz (written as {:d} Hz)".format(
            "Sample Rate", self.realSampleRate, self.sampleRate))
        print(f_int.format("Start Channel       ", self.startChannel, ""))
        print(f_int.format("Number Of Channels  ", self.numberOfChannels, ""))
        print(f_int.format("Slow Data Rate      ", self.slowDataRate,
                           "Unused"))
        print(f_int.format("Slow Start Channel  ", self.slowStartChannel,
                           "Unused"))
        print(f_int.format("Slow Num Of Channels", self.slowNumberOfChannels,
                           "Unused"))

        # Determine data type
        print(f_int.format("Data Type", self.dataType, self.dataTypeString))
        print(f_int.format("Disk Size", self.diskSize, "GB"))
        print(f_int.format("RAM Size ", self.ramSize, "MB"))
        print(f_int.format("Number Of windows", self.numberOfWindows,
                           "Unused"))


class LCDirEntry (LCCommon):
    """
    LCheapo Directory single-entry class
    """
    def __init__(self):
        pass

    def readDirEntry(self, fp):
        """
        Read an LCheapo directory entry (packed big endian format)
        """
        (self.msec, self.second, self.minute, self.hour, self.day,
         self.month, self.year) = struct.unpack('>HBBBBBB', fp.read(8))
        (self.blockNumber, self.recordLength, self.sampleRate,
         self.numBlocks) = struct.unpack('>2L2H', fp.read(12))
        (self.flag, self.muxChannel) = struct.unpack('>2B', fp.read(2))
        self.U1 = struct.unpack('>10B', fp.read(10))  # Unused

    def writeDirEntry(self, fp):
        """
        Write an LCheapo directory entry (packed big endian format)
        """
        timeData = struct.pack('>HBBBBBB', self.msec, self.second,
                               self.minute, self.hour, self.day,
                               self.month, self.year)
        flagData = struct.pack('>2L2H2B', self.blockNumber,
                               self.recordLength, self.sampleRate,
                               self.numBlocks, self.flag, self.muxChannel)
        # self.U1 doesn't work?
        uData = struct.pack('>10B', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        fp.write(timeData)
        fp.write(flagData)
        fp.write(uData)

    def __str__(self):
        """
        Create directory entry string (short format)
        """
        str_a = "{:26s} Ch:{:02d} {:7.2f}Hz     ".format(
            str(self.getDateTime()),
            self.muxChannel,
            self.getRealSampleRate(self.sampleRate))
        str_b = "{:12d} {:12d} {:12d} Flg:0x{:02x}".format(
            self.blockNumber,
            self.numBlocks,
            self.recordLength,
            self.flag)
        return str_a + str_b


class LCDataBlock (LCCommon):
    "LCheapo data block class"
    def __init__(self):
        pass

    def readBlock(self, fp):
        "Read a block of LCheapo data from the specified pointer."
        (self.msec, self.second, self.minute, self.hour, self.day,
         self.month, self.year) = struct.unpack('>HBBBBBB', fp.read(8))
        (self.blockFlag, self.muxChannel, self.numberOfSamples) =\
            struct.unpack('>BBH', fp.read(4))
        # U1=Flag=3 & U2=SampleCount=166
        (self.U1, self.U2) = struct.unpack('>BB', fp.read(2))
        self.data = fp.read(498)

    def writeBlock(self, fp):
        "Write a block of LCheapo data to the specified pointer."
        timeData = struct.pack('>HBBBBBB', self.msec, self.second,
                               self.minute, self.hour, self.day, self.month,
                               self.year)
        flagData = struct.pack('>BBH', self.blockFlag, self.muxChannel,
                               self.numberOfSamples)
        uData = struct.pack('>BB', self.U1, self.U2)
        fp.write(timeData)
        fp.write(flagData)
        fp.write(uData)
        fp.write(self.data)

    def printHexDumpOfHeader(self, annotated=False):
        "Print out the data header in hexidecimal format."
        if annotated:
            fmt = "ms:{:04x} s:{:02x} mn:{:02x} hr:{:02x} dy:{:02x} " +\
                  "mo:{:02x} yr:{:02x} Flag:{:02x} Chan:{:02x} Samples:{:04x}"
        else:
            fmt = "{:04x}{:02x}{:02x} {:02x}{:02x}{:02x}{:02x} %02x%02x%04x"
        print(fmt.format(self.msec, self.second, self.minute, self.hour,
                         self.day, self.month, self.year, self.blockFlag,
                         self.muxChannel, self.numberOfSamples))

    def printDecimalDumpOfHeader(self, annotated=False):
        "Print out the data header in decimal format."
        if annotated:
            fmt = "ms:{:4d} s:{:02d} mn:{:02d} hr:{:02d} dy:{:02d} " +\
                  "mo:{:02d} yr:{:02d} Flag:{:03d} Chan:{:02d} Samples:{:4d}"
        else:
            fmt = "{:4d} {:02d} {:02d} {:02d} {:02d} {:02d} {:02d} " +\
                  "{:03d} {:02d} {:4d}"
        print(fmt.format(self.msec, self.second, self.minute, self.hour,
                         self.day, self.month, self.year, self.blockFlag,
                         self.muxChannel, self.numberOfSamples))

    def prettyPrintHeader(self, annotated=False):
        "Print out the data header in pretty format."
        if annotated:
            fmt = "DateTime:{:02d}/{:02d}/{:02d}-{:02d}:{:02d}:{:02d}." +\
                  "{:03d} Flag:{:03d} Chan:{:02d} Samples:{:4d} " +\
                  "U1:{:03d} U2:{:03d}"
        else:
            fmt = "{:02d}/{:02d}/{:02d}-{:02d}:{:02d}:{:02d}.{:03d}" +\
                  "  F{:03d} CH{:02d} {:4d} samps U1={:03d} U2={:03d}"
        print(fmt.format(self.year, self.month, self.day, self.hour,
                         self.minute, self.second, self.msec,
                         self.blockFlag, self.muxChannel,
                         self.numberOfSamples, self.U1, self.U2))

    def convertDataTo24BitValues(self):
        "Convert the data block into a list of 24-bit values."
        data = [x * (1 << 16) + y * (1 << 8) + z
                for x, y, z in [struct.unpack(">bBB", self.data[x:x + 3])
                                for x in range(0, 498, 3)]]
        return data

    def printHexDumpOfData(self):
        "Print out the data block in hexidecimal format."
        count = 0
        for i in struct.unpack(">498B", self.data):
            sys.stdout.write("{:02x}".format(i))
            count += 1
            if count % 3 == 0:
                sys.stdout.write("  ")
            if count % 30 == 0:
                sys.stdout.write("\n")
        sys.stdout.write("\n")

    def printDecimalDumpOfData(self):
        "Print out the data block in decimal format."
        count = 0
        data = self.convertDataTo24BitValues()
        for i in data:
            sys.stdout.write("{:8d} ".format(i))
            count += 1
            if count % 8 == 0:
                sys.stdout.write("\n")
        sys.stdout.write("\n")

    def __str__(self):
        ch_str = "CH:{:d}".format(self.muxChannel)
        samp_str = "Samples:%3d".format(self.numberOfSamples)
        date_fmt = "Date:{:02d}-{:02d}-{:02d} {:02d}:{:02d}:{02d}.{:04d}"
        date_str = date_fmt.format(self.year, self.month, self.day, self.hour,
                                   self.minute, self.second, self.msec)
        return "{}  {}  {}".format(ch_str, samp_str, date_str)


def _str_from_cstr(cstr):
    """
    Convert a C string to Python
    """
    try:
        return cstr.decode("utf-8").split('\0')[0]
    except Exception:
        return None


def _cstr_from_str(str):
    """
    Convert a Python str to C string
    """
    # return str.encode("utf-8") + b'\0' Removed for comparison test
    return str.encode("utf-8")


def main():
    "Main Program"
    print("lcheapo.py is not a runnable program.")
    return 0


# Run 'main' if the script is not imported as a module)
if __name__ == '__main__':
    main()
